import os
import re
import warnings
from typing import Optional, Dict, Any, Union, Tuple, TextIO, Sequence, List

import numpy as np

from euphonic import ureg
from euphonic.util import convert_fc_phases


# h5py can't be called from Matlab, so import as late as possible to
# minimise impact. Do the same with yaml for consistency
class ImportPhonopyReaderError(ModuleNotFoundError):

    def __init__(self):
        self.message = (
            '\n\nCannot import yaml, h5py to read Phonopy files, maybe '
            'they are not installed. To install the optional '
            'dependencies for Euphonic\'s Phonopy reader, try:\n\n'
            'pip install euphonic[phonopy_reader]\n')

    def __str__(self):
        return self.message


def _convert_weights(weights: np.ndarray) -> np.ndarray:
    """
    Convert q-point weights to normalised convention

    Parameters
    ----------
    weights
        Shape (n_qpts,) float ndarray. Weights in Phonopy convention

    Returns
    ----------
    norm_weights
        Shape (n_qpts,) float ndarray. Normalised weights
    """
    total_weight = weights.sum()
    return weights/total_weight


def _extract_phonon_data_yaml(filename: str,
                              read_eigenvectors: bool = True
                              ) -> Dict[str, np.ndarray]:
    """
    From a mesh/band/qpoint.yaml file, extract the relevant information
    as a dict of Numpy arrays. No unit conversion is done at this point,
    so they are in the same units as in the .yaml file

    Parameters
    ----------
    filename
        Path and name of the mesh/band/qpoint.yaml file
    read_eigenvectors
        Whether to try and read the eigenvectors and return them in the
        dictionary

    Returns
    -------
    data_dict
        A dict with the following keys:
            'qpts', 'frequencies'
        It may also have the following keys if they are present in the
        .yaml file:
            'eigenvectors', 'weights', 'cell_vectors', 'atom_r',
            'atom_mass', 'atom_type'
    """
    try:
        import yaml
        try:
            from yaml import CSafeLoader as SafeLoader
        except ImportError:
            from yaml import SafeLoader
    except ModuleNotFoundError as e:
        raise ImportPhonopyReaderError from e

    with open(filename, 'r') as yaml_file:
        phonon_data = yaml.load(yaml_file, Loader=SafeLoader)

    data_dict = {}
    phonons = [phon for phon in phonon_data['phonon']]
    bands_data_each_qpt = [bands_data['band'] for bands_data in phonons]

    data_dict['qpts'] = np.array([phon['q-position'] for phon in phonons])
    data_dict['frequencies'] = np.array(
        [[band_data['frequency'] for band_data in bands_data]
         for bands_data in bands_data_each_qpt])
    if read_eigenvectors:
        # Eigenvectors may not be present if users haven't set
        # --eigvecs when running Phonopy - deal with this later
        try:
            data_dict['eigenvectors'] = np.squeeze(np.array(
                [[band_data['eigenvector'] for band_data in bands_data]
                 for bands_data in bands_data_each_qpt]).view(np.complex128))
        except KeyError:
            pass

    # Weights only present in mesh
    try:
        data_dict['weights'] = np.array([phon['weight'] for phon in phonons])
    except KeyError:
        pass

    # Crystal information only appears to be present in mesh/band
    try:
        data_dict['cell_vectors'] = np.array(phonon_data['lattice'])
        data_dict['atom_r'] = np.array(
            [atom['coordinates'] for atom in phonon_data['points']])
        data_dict['atom_mass'] = np.array(
            [atom['mass'] for atom in phonon_data['points']])
        data_dict['atom_type'] = np.array(
            [atom['symbol'] for atom in phonon_data['points']])
    except KeyError:
        pass
    return data_dict


def _extract_phonon_data_hdf5(filename: str,
                              read_eigenvectors: bool = True
                              ) -> Dict[str, np.ndarray]:
    """
    From a mesh/band/qpoint.hdf5 file, extract the relevant information
    as a dict of Numpy arrays. No unit conversion is done at this point,
    so they are in the same units as in the .hdf5 file

    Parameters
    ----------
    filename
        Path and name of the mesh/band/qpoint.hdf5 file
    read_eigenvectors
        Whether to try and read the eigenvectors and return them in the
        dictionary

    Returns
    -------
    data_dict
        A dict with the following keys:
            'qpts', 'frequencies'
        It may also have the following keys if they are present in the
        .hdf5 file:
            'eigenvectors', 'weights'
    """
    try:
        import h5py
    except ModuleNotFoundError as e:
        raise ImportPhonopyReaderError from e

    with h5py.File(filename, 'r') as hdf5_file:
        data_dict = {}
        if 'qpoint' in hdf5_file:
            data_dict['qpts'] = hdf5_file['qpoint'][()]
            data_dict['frequencies'] = hdf5_file['frequency'][()]
            if read_eigenvectors:
                # Eigenvectors may not be present if users haven't set
                # --eigvecs when running Phonopy - deal with this later
                try:
                    data_dict['eigenvectors'] = hdf5_file['eigenvector'][()]
                except KeyError:
                    pass
            # Only mesh.hdf5 has weights
            try:
                data_dict['weights'] = hdf5_file['weight'][()]
            except KeyError:
                pass
        # Is a band.hdf5 file - q-points are stored in 'path' and need
        # special treatment
        else:
            data_dict['qpts'] = hdf5_file['path'][()].reshape(
                -1, hdf5_file['path'][()].shape[-1])
            data_dict['frequencies'] = hdf5_file['frequency'][()].reshape(
                -1, hdf5_file['frequency'][()].shape[-1])
            try:
                # The last 2 dimensions of eigenvectors in bands.hdf5 are for
                # some reason transposed compared to mesh/qpoints.hdf5, so also
                # transpose to handle this
                data_dict['eigenvectors'] = hdf5_file[
                    'eigenvector'][()].reshape(
                        -1, *hdf5_file['eigenvector'][()].shape[-2:]
                        ).transpose([0,2,1])
            except KeyError:
                pass

    return data_dict

def read_phonon_data(
        path: str = '.',
        phonon_name: str = 'band.yaml',
        phonon_format: Optional[str] = None,
        summary_name: str = 'phonopy.yaml',
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        read_eigenvectors: bool = True
        ) -> Dict[str, Union[int, str, np.ndarray]]:
    """
    Reads precalculated phonon mode data from a Phonopy
    mesh/band/qpoints.yaml/hdf5 file and returns it in a dictionary.
    May also read from phonopy.yaml for structure information. The
    output eigenvectors will have their phase transformed from
    per atom to per cell phases.

    Parameters
    ----------
    path
        Path to directory containing the file(s)
    phonon_name
        Name of Phonopy file including the frequencies and eigenvectors
    phonon_format
        Format of the phonon_name file if it isn't obvious from the
        phonon_name extension. One of {'yaml', 'hdf5'}
    summary_name
        Name of Phonopy summary file to read the crystal information
        from. Crystal information in the phonon_name file takes
        priority, but if it isn't present, crystal information is read
        from summary_name instead
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in
    read_eigenvectors
        Whether to read the eigenvectors and return them in the
        dictionary. In this case, if eigenvectors = False it is assumed
        a QpointFrequencies object is being created so detailed structure
        information such as atom locations, masses, species is not
        explicitly required, although the cell vectors are required.

    Returns
    -------
    data_dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit'.

        If read_eigenvectors is True, there will also be an
        'eigenvectors' key
    """
    phonon_pathname = os.path.join(path, phonon_name)
    summary_pathname = os.path.join(path, summary_name)

    hdf5_exts = ['hdf5', 'hd5', 'h5']
    yaml_exts = ['yaml', 'yml', 'yl']
    if phonon_format is None:
        phonon_format = os.path.splitext(phonon_name)[1].strip('.')

    if phonon_format in hdf5_exts:
        phonon_dict = _extract_phonon_data_hdf5(
            phonon_pathname, read_eigenvectors=read_eigenvectors)
    elif phonon_format in yaml_exts:
        phonon_dict = _extract_phonon_data_yaml(
            phonon_pathname, read_eigenvectors=read_eigenvectors)
    else:
        raise ValueError((f'File format {phonon_format} of {phonon_name}'
                          f' is not recognised'))

    if read_eigenvectors and not 'eigenvectors' in phonon_dict:
        raise RuntimeError((f'Eigenvectors couldn\'t be found in '
                            f'{phonon_pathname}, ensure --eigvecs was '
                            f'set when running Phonopy'))

    # Since units are not explicitly defined in
    # mesh/band/qpoints.yaml/hdf5 assume:
    ulength = 'angstrom'
    umass = 'amu'
    ufreq = 'THz'

    crystal_keys = ['cell_vectors', 'atom_r', 'atom_mass', 'atom_type']
    # Check if crystal structure has been read from phonon_file, if not
    # get structure from summary_file
    if len(crystal_keys & phonon_dict.keys()) != len(crystal_keys):
        summary_dict = _extract_summary(summary_pathname)
        phonon_dict['cell_vectors'] = summary_dict['cell_vectors']
        phonon_dict['atom_r'] = summary_dict['atom_r']
        phonon_dict['atom_mass'] = summary_dict['atom_mass']
        phonon_dict['atom_type'] = summary_dict['atom_type']
        # Overwrite assumed units if they are found in summary file
        ulength = summary_dict['ulength']
        umass = summary_dict['umass']
        # Check phonon_file and summary_file are commensurate
        if 3*len(phonon_dict['atom_r']) != len(phonon_dict['frequencies'][0]):
            raise ValueError((
                f'Phonon file {phonon_pathname} not commensurate '
                f'with summary file {summary_pathname}. Please '
                'check contents'))

    data_dict: Dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['n_atoms'] = len(phonon_dict['atom_r'])
    cry_dict['cell_vectors'] = phonon_dict['cell_vectors']*ureg(
        ulength).to(cell_vectors_unit).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = phonon_dict['atom_r']
    cry_dict['atom_type'] = phonon_dict['atom_type']
    cry_dict['atom_mass'] = phonon_dict['atom_mass']*ureg(
        umass).to(atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit
    n_qpts = len(phonon_dict['qpts'])
    data_dict['n_qpts'] = n_qpts
    data_dict['qpts'] = phonon_dict['qpts']
    data_dict['frequencies'] = phonon_dict['frequencies']*ureg(
        ufreq).to(frequencies_unit).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    if read_eigenvectors:
        # Convert Phonopy conventions to Euphonic conventions
        data_dict['eigenvectors'] = convert_eigenvector_phases(phonon_dict)
    if 'weights' in phonon_dict:
        data_dict['weights'] = _convert_weights(phonon_dict['weights'])
    return data_dict


def convert_eigenvector_phases(phonon_dict: Dict[str, np.ndarray]
                               ) -> np.ndarray:
    """
    When interpolating the force constants matrix, Euphonic uses a phase
    convention of e^iq.r_a, where r_a is the coordinate of each CELL in
    the supercell, whereas Phonopy uses e^iq.r_k, where r_k is the
    coordinate of each ATOM in the supercell. This must be accounted for
    when reading Phonopy eigenvectors by applying a phase of
    e^-iq(r_k - r_k')
    """
    atom_r = phonon_dict['atom_r']
    n_atoms = len(atom_r)
    qpts = phonon_dict['qpts']
    n_qpts = len(qpts)

    eigvecs = np.reshape(phonon_dict['eigenvectors'],
                         (n_qpts, n_atoms, 3, n_atoms, 3))
    na = np.newaxis
    rk_diff = atom_r[:, na, :] - atom_r[na, :, :]
    conversion = np.exp(-2j*np.pi*np.einsum('il,jkl->ijk', qpts, rk_diff))
    eigvecs = np.einsum('ijklm,ijl->ijklm', eigvecs, conversion)
    return np.reshape(eigvecs, (n_qpts, 3*n_atoms, n_atoms, 3))


def _extract_force_constants(fc_pathname: str, n_atoms: int, n_cells: int,
                             summary_name: str) -> np.ndarray:
    """
    Reads force constants from a Phonopy FORCE_CONSTANTS file

    Parameters
    ----------
    fc_pathname
        The FORCE_CONSTANTS file to read from
    n_atoms
        Number of atoms in the unit cell
    n_cells
        Number of unit cells in the supercell
    summary_name
        Name of input phonopy.yaml file

    Returns
    -------
    fc
        Shape (n_atoms, n_atoms*n_cells, 3, 3) or
        (n_atoms*n_cells, n_atoms*n_cells, 3, 3) float ndarray. The
        force constants in Phonopy convention
    """
    with open(fc_pathname, 'r') as f:
        fc_dims =  [int(dim) for dim in f.readline().split()]
    # single shape specifier implies full format
    if len(fc_dims) == 1:
        fc_dims.append(fc_dims[0])
    _check_fc_shape(fc_dims, n_atoms, n_cells, fc_pathname, summary_name)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        fc = np.genfromtxt(fc_pathname, skip_header=1,
                           max_rows=4*(n_atoms*n_cells)**2, usecols=(0,1,2),
                           invalid_raise=False)
    return fc.reshape(fc_dims + [3, 3])


def _extract_force_constants_hdf5(
        filename: str, n_atoms: int, n_cells: int, summary_name: str,
        ) -> np.ndarray:
    try:
        import h5py
    except ModuleNotFoundError as e:
        raise ImportPhonopyReaderError from e

    with h5py.File(filename, 'r') as fc_file:
        fc = fc_file['force_constants'][:]
        _check_fc_shape(fc.shape, n_atoms, n_cells, fc_file.filename,
                        summary_name)
    return fc


def _check_fc_shape(fc_shape: Tuple[int, int], n_atoms: int,
                    n_cells: int, fc_filename: str,
                    summary_filename: str) -> None:
    """
    Check if force constants has the correct shape
    """
    if (not ((fc_shape[0] == n_atoms or fc_shape[0] == n_cells*n_atoms) and
             fc_shape[1] == n_cells*n_atoms)):
        raise ValueError((
            f'Force constants matrix with shape {fc_shape} read from '
            f'{fc_filename} is not compatible with crystal read from '
            f'{summary_filename} which has {n_atoms} atoms in the cell,'
            f' and {n_cells} cells in the supercell'))


def _extract_born(born_file_obj: TextIO) -> Dict[str, Union[float, np.ndarray]]:
    """
    Parse and convert dielectric tensor and born effective
    charge from BORN file

    Parameters
    ----------
    born_file_obj
        Pointer to BORN file

    Returns
    ----------
    born_dict
        Dict containing dielectric tensor and born effective charge
    """
    born_lines_str = born_file_obj.readlines()
    born_lines: List[List[float]] = []
    for line in born_lines_str:
        # Ignore comments
        if line.startswith('#'):
            pass
        else:
            born_lines.append([float(x) for x in line.split()])

    born_dict: Dict[str, Union[float, np.ndarray]] = {}

    idx0 = 0
    if len(born_lines[0]) == 1:
        # Then this is the NAC conversion factor
        born_dict['nac_factor'] = born_lines[0][0]
        idx0 = 1

    # dielectric first line after factor
    # xx, xy, xz, yx, yy, yz, zx, zy, zz.
    born_dict['dielectric'] = np.array(born_lines[idx0]).reshape([3,3])

    # born charges after dielectric
    # xx, xy, xz, yx, yy, yz, zx, zy, zz.
    born_dict['born'] = np.array(
        [np.array(bl).reshape([3,3]) for bl in born_lines[idx0+1:]])

    return born_dict


def _extract_summary(filename: str, fc_extract: bool = False
                     ) -> Dict[str, Union[str, int, np.ndarray]]:
    """
    Read phonopy.yaml for summary data produced during the Phonopy
    post-process

    Parameters
    ----------
    filename
        Path and name of the Phonopy summary file (usually phonopy.yaml)
    fc_extract
        Whether to attempt to read force constants and related
        information from summary_object

    Returns
    ----------
    summary_dict
        A dict with the following keys: n_atoms, cell_vectors, atom_r,
        atom_type, atom_mass, ulength, umass. Also optionally has the
        following keys: sc_matrix, n_cells_in_sc, cell_origins,
        cell_origins_map, force_constants, ufc, born, dielectric
    """
    try:
        import yaml
        try:
            from yaml import CSafeLoader as SafeLoader
        except ImportError:
            from yaml import SafeLoader
    except ModuleNotFoundError as e:
        raise ImportPhonopyReaderError from e

    with open(filename, 'r') as summary_file:
        summary_object = yaml.load(summary_file, Loader=SafeLoader)
    (cell_vectors, n_atoms, atom_r, atom_mass,
     atom_type, _) = _extract_crystal_data(summary_object['primitive_cell'])

    summary_dict = {}
    try:
        pu = summary_object['physical_unit']
    except KeyError:
        default_units = {'atomic_mass': 'AMU',
                         'length': 'Angstrom',
                         'force_constants': 'eV/Angstrom^2'}
        print(f'physical_unit key not found in {filename}, assuming '
              f'the following units: {default_units}')
        pu = default_units
    # Format units so they can be read by Pint
    pu['atomic_mass'] = pu['atomic_mass'].replace('AMU', 'amu')
    # Ensure denominator is surrounded by brackets to handle
    # cases like 'eV/Angstrom.au'
    divs = re.findall(r'/([^/]+)', pu['force_constants'])
    pu['force_constants'] = (pu['force_constants'].split('/')[0]
                             + '/(' + '/('.join([d+')' for d in divs]))
    for key, value in pu.items():
        pu[key] = value.replace('au', 'bohr').replace('Angstrom', 'angstrom')

    summary_dict['ulength'] = pu['length']
    summary_dict['umass'] = pu['atomic_mass']

    summary_dict['n_atoms'] = n_atoms
    summary_dict['cell_vectors'] = cell_vectors
    summary_dict['atom_r'] = atom_r
    summary_dict['atom_type'] = atom_type
    summary_dict['atom_mass'] = atom_mass

    if fc_extract:
        # Get matrix to convert from primitive cell to unit cell
        u_to_sc_matrix = np.array(summary_object['supercell_matrix'])
        if 'primitive_matrix' in summary_object:
            u_to_p_matrix = np.array(summary_object['primitive_matrix'])
            # Matrix to convert from primitive to supercell
            p_to_u_matrix = np.linalg.inv(u_to_p_matrix).transpose()
            p_to_sc_matrix = np.rint(
                np.matmul(u_to_sc_matrix, p_to_u_matrix)).astype(np.int32)
        else:
            u_to_p_matrix = np.identity(3, dtype=np.int32)
            p_to_sc_matrix = u_to_sc_matrix

        # Get supercell info
        _, _, satom_r, _, _, sc_idx_in_pc = _extract_crystal_data(
            summary_object['supercell'])
        # Coords of supercell atoms in frac coords of the prim cell
        satom_r_pcell = np.einsum('ij,jk->ik', satom_r, p_to_sc_matrix)
        # Determine mapping from atoms in the supercell to the prim cell
        # Note: -1 to get 0-indexed instead of 1-indexed values
        pc_to_sc_atom_idx, sc_to_pc_atom_idx = np.unique(
            sc_idx_in_pc - 1, return_inverse=True)

        summary_dict['sc_matrix'] = p_to_sc_matrix
        summary_dict['sc_atom_r'] = satom_r_pcell
        summary_dict['pc_to_sc_atom_idx'] = pc_to_sc_atom_idx
        summary_dict['sc_to_pc_atom_idx'] = sc_to_pc_atom_idx
        summary_dict['ufc'] = pu['force_constants']
        try:
            fc = summary_object['force_constants']
            summary_dict['force_constants'] = np.array(
                fc['elements']).reshape(fc['shape'] + [3,3])
        except KeyError:
            pass
        # NAC factor may be present even if born is not in phonopy.yaml
        # (born may be in separate BORN file), and the BORN file may
        # not necessarily contain NAC factor, so just try reading it
        # from phonopy.yaml anyway
        try:
            summary_dict['nac_factor'] = summary_object[
                'phonopy']['nac_unit_conversion_factor']
        except KeyError:
            pass
        try:
            summary_dict['born'] = np.array(
                summary_object['born_effective_charge'])
            summary_dict['dielectric'] = np.array(
                summary_object['dielectric_constant'])
        except KeyError:
            pass

    return summary_dict


def _extract_crystal_data(crystal: Dict[str, Any]
                          ) -> Tuple[np.ndarray, int, np.ndarray,
                                     np.ndarray, np.ndarray, np.ndarray]:
    """
    Gets relevant data from a section of phonopy.yaml

    Parameters
    ----------
    crystal
        Part of the dict obtained from reading a phonopy.yaml file. e.g.
        summary_dict['unit_cell']

    Returns
    -------
    cell_vectors
        Shape (3,3) float ndarray. Cell vectors, in same units as in
        the phonopy.yaml file
    n_atoms
        Number of atoms
    atom_r
        Shape (n_atoms, 3) float ndarray. Fractional position of each
        atom
    atom_mass
        Shape (n_atoms,) float ndarray. Mass of each atom, in same
        units as in the phonopy.yaml file
    atom_type
        Shape (n_atoms,) str ndarray. String specifying the species
        of each atom
    idx_in_pcell
        Shape (n_atoms,) int ndarray. Maps the atom index on the
        unit/supercell to the primitive cell
    """
    n_atoms = len(crystal['points'])
    cell_vectors = np.array(crystal['lattice'])
    atom_r = np.zeros((n_atoms, 3))
    atom_mass = np.zeros(n_atoms)
    atom_type = np.array([])
    for i in range(n_atoms):
        atom_mass[i] = crystal['points'][i]['mass']
        atom_r[i] = crystal['points'][i]['coordinates']
        atom_type = np.append(atom_type, crystal['points'][i]['symbol'])

    if 'reduced_to' in crystal['points'][0]:
        idx_in_pcell = np.zeros(n_atoms, dtype=np.int32)
        for i in range(n_atoms):
            idx_in_pcell[i] = crystal['points'][i]['reduced_to']
    else:
        # If reduced_to isn't present it is already the primitive cell
        idx_in_pcell = np.arange(n_atoms, dtype=np.int32)

    return cell_vectors, n_atoms, atom_r, atom_mass, atom_type, idx_in_pcell


def read_interpolation_data(
        path: str = '.',
        summary_name: str = 'phonopy.yaml',
        born_name: Optional[str] = None,
        fc_name: str = 'FORCE_CONSTANTS',
        fc_format: Optional[str] = None,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        force_constants_unit: str = 'hartree/bohr**2',
        born_unit: str = 'e',
        dielectric_unit: str = '(e**2)/(bohr*hartree)') -> Dict[str, Any]:
    """
    Reads data from the phonopy summary file (default phonopy.yaml) and
    optionally born and force constants files. Only attempts to read
    from born or force constants files if these can't be found in the
    summary file. Note that the output force constants will be transformed
    to a Euphonic-like shape, but if it is a polar material it will still
    be the long-ranged force constants matrix (and not the short-ranged one
    that Euphonic requires in the case of polar materials).

    Parameters
    ----------
    path
        Path to directory containing the file(s)
    summary_name
        Filename of phonopy summary file, default phonopy.yaml. By
        default any information (e.g. force constants) read from this
        file takes priority
    born_name
        Name of the Phonopy file containing born charges and dielectric
        tensor, (by convention in Phonopy this would be called BORN). Is
        only read if Born charges can't be found in the summary_name
        file
    fc_name
        Name of file containing force constants. Is only read if force
        constants can't be found in summary_name
    fc_format
        One of {'phonopy', 'hdf5'}. Format of file containing force
        constants data. FORCE_CONSTANTS is type 'phonopy'
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    force_constants_unit
        The unit to return the force constants in
    born_unit
        The unit to return the Born charges in
    dielectric_unit
        The unit to return the dielectric permittivity tensor in

    Returns
    -------
    data_dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'force_constants', 'force_constants_unit',
        'sc_matrix' and 'cell_origins'. Also contains 'born',
        'born_unit', 'dielectric' and 'dielectric_unit' if they are
        present in phonopy.yaml
    """
    summary_pathname = os.path.join(path, summary_name)
    summary_dict = _extract_summary(summary_pathname, fc_extract=True)

    # Only read force constants if it's not in summary file
    if not 'force_constants' in summary_dict:
        hdf5_exts = ['hdf5', 'hd5', 'h5']
        if fc_format is None:
            fc_format = os.path.splitext(fc_name)[1].strip('.')
            if fc_format not in hdf5_exts:
                fc_format = 'phonopy'
        fc_pathname = os.path.join(path, fc_name)
        print((f'Force constants not found in {summary_pathname}, '
               f'attempting to read from {fc_pathname}'))
        n_atoms = summary_dict['n_atoms']
        n_cells = int(len(summary_dict['sc_atom_r'])/n_atoms)
        if fc_format == 'phonopy':
            with open(fc_pathname, 'r') as fc_file:
                summary_dict['force_constants'] = _extract_force_constants(
                    fc_pathname, n_atoms, n_cells, summary_pathname)
        elif fc_format in 'hdf5':
            summary_dict['force_constants'] =  _extract_force_constants_hdf5(
                fc_pathname, n_atoms, n_cells, summary_pathname)
        else:
            raise ValueError((
                f'Force constants file format "{fc_format}" of '
                f'"{fc_name}" is not recognised'))

    # Only read born/dielectric if they're not in summary file and the
    # user has specified a Born file
    dipole_keys = ['born', 'dielectric', 'nac_factor']
    if (born_name is not None and
            len(dipole_keys & summary_dict.keys()) != len(dipole_keys)):
        born_pathname = os.path.join(path, born_name)
        print((f'Born, dielectric not found in {summary_pathname}, '
               f'attempting to read from {born_pathname}'))
        with open(born_pathname, 'r') as born_file:
            born_dict = _extract_born(born_file)
        # Let BORN file take priority, but merge because the 'nac_factor'
        # key may not always be present in BORN
        summary_dict.update(born_dict)
    # Check if born key is present, then factor is also present. It
    # may not always be e.g. you can run Phonopy so that 'born',
    # 'dielectric' are written to phonopy.yaml, but if NAC = .FALSE.
    # 'nac_factor' will not be written. In this case raise error.
    if ('born' in summary_dict
            and 'nac_factor' not in summary_dict):
        raise KeyError(f'nac_unit_conversion_factor could not be found in '
                       f'{summary_pathname} or the BORN file (if given), so '
                       f'units of the dielectric tensor cannot be determined.')

    # Units from summary_file
    ulength = summary_dict['ulength']
    umass = summary_dict['umass']
    ufc = summary_dict['ufc']

    data_dict: Dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['cell_vectors'] = summary_dict['cell_vectors']*ureg(
        ulength).to(cell_vectors_unit).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    # Normalise atom coordinates
    cry_dict['atom_r'] = (summary_dict['atom_r']
                          - np.floor(summary_dict['atom_r']))
    cry_dict['atom_type'] = summary_dict['atom_type']
    cry_dict['atom_mass'] = summary_dict['atom_mass']*ureg(
        umass).to(atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    fc, cell_origins = convert_fc_phases(
         summary_dict['force_constants'], summary_dict['atom_r'],
         summary_dict['sc_atom_r'], summary_dict['pc_to_sc_atom_idx'],
         summary_dict['sc_to_pc_atom_idx'], summary_dict['sc_matrix'])
    data_dict['force_constants'] = fc*ureg(
        ufc).to(force_constants_unit).magnitude
    data_dict['force_constants_unit'] = force_constants_unit
    data_dict['sc_matrix'] = summary_dict['sc_matrix']
    data_dict['cell_origins'] = cell_origins

    try:
        data_dict['born'] = summary_dict['born']*ureg(
            'e').to(born_unit).magnitude
        data_dict['born_unit'] = born_unit
        # Phonopy NAC conversion factor converts the following NAC
        # term (eq. 75 in Gonze & Lee 1997, ignoring dimensionless
        # units)
        # born**2/(volume x dielectric)
        #     = 1/(ulength**3 x di_energy x di_length)
        # to the same units as the force constants (given by ufc)
        # i.e. factor*nac_units = force constants units
        # So allow pint to do this unit conversion automatically
        phonopy_dielectric_unit = ureg('e**2')/(ureg(ulength)**3*ureg(ufc))
        data_dict['dielectric'] = (
            summary_dict['dielectric']/summary_dict['nac_factor']
            *phonopy_dielectric_unit.to(dielectric_unit).magnitude)
        data_dict['dielectric_unit'] = dielectric_unit
    except KeyError:
        pass

    return data_dict



import warnings
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Type

import numpy as np

from euphonic.validate import _check_constructor_inputs, _check_unit_conversion
from euphonic.io import (_obj_to_json_file, _obj_from_json_file,
                         _obj_to_dict, _process_dict)
from euphonic.readers import castep, phonopy
from euphonic.util import (_calc_abscissa, get_qpoint_labels)
from euphonic import (ureg, Crystal, Quantity, Spectrum1D,
                      Spectrum1DCollection, Spectrum2D)
from euphonic.broadening import _width_interpolated_broadening


class QpointFrequencies:
    """
    A class to read and store frequency data at q-points

    Attributes
    ----------
    crystal
        Lattice and atom information
    n_qpts
        Number of q-points in the object
    qpts
        Shape (n_qpts, 3) float ndarray. Q-point coordinates, in
        fractional coordinates of the reciprocal lattice
    frequencies
        Shape (n_qpts, n_branches) float Quantity in energy units.
        Frequencies per q-point and mode
    weights
        Shape (n_qpts,) float ndarray. The weight for each q-point
    """
    T = TypeVar('T', bound='QpointFrequencies')

    def __init__(self, crystal: Crystal, qpts: np.ndarray,
                 frequencies: Quantity,
                 weights: Optional[np.ndarray] = None) -> None:
        """
        Parameters
        ----------
        crystal
            Lattice and atom information
        qpts
            Shape (n_qpts, 3) float ndarray. Q-point coordinates
        frequencies
            Shape (n_qpts, n_branches) float Quantity in energy units.
            Frequencies per q-point and mode
        weights
            Shape (n_qpts,) float ndarray. The weight for each q-point.
            If None, equal weights are assumed
        """
        _check_constructor_inputs(
            [crystal, qpts], [Crystal, np.ndarray], [(), (-1, 3)],
            ['crystal', 'qpts'])
        n_qpts = len(qpts)
        # Unlike QpointPhononModes and StructureFactor, don't test the
        # frequencies shape against number of atoms in the crystal, as
        # we may only have the cell vectors
        _check_constructor_inputs(
            [frequencies, weights],
            [Quantity, [np.ndarray, type(None)]],
            [(n_qpts, -1), (n_qpts,)],
            ['frequencies', 'weights'])
        self.crystal = crystal
        self.qpts = qpts
        self.n_qpts = n_qpts
        self._frequencies = frequencies.to(ureg.hartree).magnitude
        self.frequencies_unit = str(frequencies.units)

        if weights is not None:
            self.weights = weights
        else:
            self.weights = np.full(self.n_qpts, 1/self.n_qpts)

    @property
    def frequencies(self) -> Quantity:
        return self._frequencies*ureg('hartree').to(self.frequencies_unit)

    @frequencies.setter
    def frequencies(self, value: Quantity) -> None:
        self.frequencies_unit = str(value.units)
        self._frequencies = value.to('hartree').magnitude

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['frequencies_unit'])
        super(QpointFrequencies, self).__setattr__(name, value)

    def calculate_dos(self, dos_bins: Quantity,
                      mode_widths: Optional[Quantity] = None,
                      mode_widths_min: Quantity = Quantity(0.01, 'meV'),
                      adaptive_method: str = 'reference',
                      adaptive_error: float = 0.01,
                      ) -> Spectrum1D:
        """
        Calculates a density of states, in units of modes per atom per
        energy unit, such that the integrated area is equal to 3.

        Parameters
        ----------
        dos_bins
            Shape (n_e_bins + 1,) float Quantity in energy units. The
            energy bin edges to use for calculating the DOS
        mode_widths
            Shape (n_qpts, n_branches) float Quantity in energy units.
            The broadening width for each mode at each q-point, for
            adaptive broadening
        mode_widths_min
            Scalar float Quantity in energy units. Sets a lower limit on
            the mode widths, as mode widths of zero will result in
            infinitely sharp peaks
        adaptive_method
            String. Specifies whether to use slow, reference adaptive method or
            faster, approximate method. Allowed options are 'reference'
            or 'fast', default is 'reference'.
        adaptive_error
            Scalar float. Acceptable error for gaussian approximations
            when using the fast adaptive method, defined as the absolute
            difference between the areas of the true and approximate gaussians

        Returns
        -------
        dos
            A spectrum containing the energy bins on the x-axis and DOS
            on the y-axis. The DOS is in units of modes per energy unit
            per atom, such that the integrated area is equal to 3.

        Notes
        -----

        The fast adaptive broadening method reduces computation time by
        reducing the number of Gaussian functions that have to be evaluated.
        Broadening kernels are only evaulated for a subset of mode width values
        with intermediate values approximated using interpolation.

        The ``adaptive_error`` keyword argument is used to determine how many
        broadening kernels are computed exactly. The more exact kernels are
        used, the more accurate the Gaussian approximations will be, but
        computation time will also be increased.

        """
        dos = self._calculate_dos(dos_bins, mode_widths=mode_widths,
                                  mode_widths_min=mode_widths_min,
                                  adaptive_method=adaptive_method,
                                  adaptive_error=adaptive_error)
        return Spectrum1D(dos_bins, dos)

    def _calculate_dos(self, dos_bins: Quantity,
                       mode_widths: Optional[Quantity] = None,
                       mode_widths_min: Quantity = Quantity(0.01, 'meV'),
                       mode_weights: Optional[np.ndarray] = None,
                       adaptive_method: str = 'reference',
                       adaptive_error: float = 0.01,
                       q_idx: Optional[int] = None
                       ) -> Quantity:
        """
        Calculates a density of states (same arg defs as calculate_dos),
        but returns just a Quantity containing the y_data, rather
        than a Spectrum1D

        The mode_weights is a shape (n_qpts, n_modes) float array,
        describing the weight of each mode at each q-point.
        This is useful for calculating PDOS, for total DOS each
        mode is equally weighted so each mode_weights value will
        be one, for PDOS it will not

        If q_idx is specified, only calculates a DOS for that specific
        q-point, used for calculating a Q-E DOS-weighted map. In this
        case, mode_widths and mode_weights must be shape (1, n_modes)
        arrays
        """

        adaptive_method_options = ['reference', 'fast']
        if adaptive_method not in adaptive_method_options:
            raise ValueError(f'Invalid value for adaptive_method, '
                             f'got {adaptive_method}, should be one '
                             f'of {adaptive_method_options}')

        if q_idx is None:
            freqs = self._frequencies
            weights = self.weights
        else:
            freqs = self._frequencies[q_idx][np.newaxis, :]
            weights = np.array([self.weights[q_idx]])
        n_modes = freqs.shape[-1]
        # dos_bins commonly contains a 0 bin, and converting from 0 1/cm
        # to 0 hartree causes a RuntimeWarning, so suppress it
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            dos_bins_calc = dos_bins.to('hartree').magnitude
        if mode_weights is not None:
            mode_weights_calc = mode_weights
        else:
            # Normalise DOS to per atom (rather than per unit cell)
            mode_weights_calc = np.full(freqs.shape, 3/freqs.shape[1])
        if mode_widths is not None:
            mode_widths = mode_widths.to('hartree').magnitude
            mode_widths = np.maximum(mode_widths,
                                     mode_widths_min.to('hartree').magnitude)
            if adaptive_method == 'reference':
                # adaptive broadening by summing over individual peaks
                from scipy.stats import norm
                dos_bins_calc = Spectrum1D._bin_edges_to_centres(dos_bins_calc)
                dos = np.zeros(len(dos_bins_calc))
                for q in range(len(freqs)):
                    for m in range(n_modes):
                        pdf = norm.pdf(dos_bins_calc, loc=freqs[q, m],
                                       scale=mode_widths[q, m])
                        dos += pdf*weights[q]*mode_weights_calc[q, m]
            elif adaptive_method == 'fast':
                # fast, approximate method for adaptive broadening
                combined_weights = mode_weights_calc * weights[:, np.newaxis]
                dos = _width_interpolated_broadening(dos_bins_calc,
                                                     freqs, mode_widths,
                                                     combined_weights,
                                                     adaptive_error)
        else:
            bin_idx = np.digitize(freqs, dos_bins_calc)
            # Create DOS with extra bin either side, for any points
            # that fall outside the bin range
            dos = np.zeros(len(dos_bins) + 1)
            bin_widths = np.ones(len(dos_bins) + 1)  # Use ones to avoid div/0
            bin_widths[1:-1] = np.diff(dos_bins_calc)
            mode_weights_calc = (weights[:, np.newaxis]
                                 * mode_weights_calc/bin_widths[bin_idx])
            np.add.at(dos, bin_idx, mode_weights_calc)
            dos = dos[1:-1]

        dos = dos/np.sum(weights)
        # Avoid issues in converting DOS when energy is in cm^-1
        # Pint allows hartree -> cm^-1 but not 1/hartree -> cm
        conv = 1*ureg('hartree').to(dos_bins.units)
        return dos/conv

    def calculate_dos_map(self, dos_bins: Quantity,
                          mode_widths: Optional[Quantity] = None,
                          mode_widths_min: Quantity = Quantity(0.01, 'meV')
                          ) -> Spectrum2D:
        """
        Produces a bandstructure-like plot, using the DOS at each q-point

        Parameters
        ----------
        dos_bins
            Shape (n_e_bins + 1,) float Quantity in energy units. The
            energy bin edges to use for calculating the DOS
        mode_widths
            Shape (n_qpts, n_branches) float Quantity in energy units.
            The broadening width for each mode at each q-point, for
            adaptive broadening
        mode_widths_min
            Scalar float Quantity in energy units. Sets a lower limit on
            the mode widths, as mode widths of zero will result in
            infinitely sharp peaks

        Returns
        -------
        dos_map
            A 2D spectrum containing the q-point bins on the x-axis, energy
            bins on the y-axis and DOS on the z-axis
        """
        x_data, x_tick_labels = self._get_qpt_axis_and_labels()
        dos_map = np.zeros((len(self.qpts), len(dos_bins) - 1))
        for i in range(len(self.qpts)):
            dos_kwargs = {}
            if mode_widths is not None:
                dos_kwargs['mode_widths'] = mode_widths[i][np.newaxis, :]
            dos = self._calculate_dos(
                dos_bins, mode_widths_min=mode_widths_min, q_idx=i,
                **dos_kwargs)
            dos_map[i, :] = dos.magnitude
        return Spectrum2D(x_data, dos_bins, dos_map*(dos.units),
                          x_tick_labels=x_tick_labels)

    def get_dispersion(self) -> Spectrum1DCollection:
        """
        Creates a set of 1-D bands from mode data

        Bands follow the same q-point order as in the qpts array, with
        x-axis spacing corresponding to the absolute distances between
        q-points.  Discontinuities will appear as large jumps on the
        x-axis.

        Returns
        -------
        dispersion
            A sequence of mode bands with a common x-axis
        """
        abscissa = _calc_abscissa(self.crystal.reciprocal_cell(), self.qpts)
        x_tick_labels = get_qpoint_labels(self.qpts,
                                          cell=self.crystal.to_spglib_cell())
        return Spectrum1DCollection(abscissa, self.frequencies.T,
                                    x_tick_labels=x_tick_labels)

    def _get_qpt_axis_and_labels(
            self) -> Tuple[Quantity, List[Tuple[int, str]]]:
        """
        Converts the qpts stored in this object an array of
        scalar distances and applies appropriate symmetry labels

        Returns
        -------
        abscissa
            Shape (n_qpts,) float quantity in 1/length units. The
            scalar q-point distances
        x_tick_labels
            The tick labels
        """
        abscissa = _calc_abscissa(self.crystal.reciprocal_cell(), self.qpts)
        # Calculate q-space ticks and labels
        x_tick_labels = get_qpoint_labels(
            self.qpts, cell=self.crystal.to_spglib_cell())
        return abscissa, x_tick_labels

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary. See QpointFrequencies.from_dict for
        details on keys/values
        """
        dout = _obj_to_dict(self, ['crystal', 'n_qpts', 'qpts', 'frequencies',
                                   'weights'])
        return dout

    def to_json_file(self, filename: str) -> None:
        """
        Write to a JSON file. JSON fields are equivalent to
        from_dict keys

        Parameters
        ----------
        filename
            Name of the JSON file to write to
        """
        _obj_to_json_file(self, filename)

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """
        Convert a dictionary to a QpointFrequencies object

        Parameters
        ----------
        d
            A dictionary with the following keys/values:

            - 'crystal': dict, see Crystal.from_dict
            - 'qpts': (n_qpts, 3) float ndarray
            - 'frequencies': (n_qpts, n_branches) float ndarray
            - 'frequencies_unit': str

            There are also the following optional keys:

            - 'weights': (n_qpts,) float ndarray
        """
        crystal = Crystal.from_dict(d['crystal'])
        d = _process_dict(d, quantities=['frequencies'], optional=['weights'])
        return cls(crystal, d['qpts'], d['frequencies'],
                   d['weights'])

    @classmethod
    def from_json_file(cls: Type[T], filename: str) -> T:
        """
        Read from a JSON file. See from_dict for
        required fields

        Parameters
        ----------
        filename
            The file to read from
        """
        return _obj_from_json_file(cls, filename)

    @classmethod
    def from_castep(cls: Type[T], filename: str,
                    average_repeat_points: bool = True) -> T:
        """
        Reads precalculated phonon mode data from a CASTEP .phonon file

        Parameters
        ----------
        filename
            The path and name of the .phonon file to read
        average_repeat_points
            If multiple frequency/eigenvectors blocks are included with the
            same q-point index (i.e. for Gamma-point with LO-TO splitting),
            scale the weights such that these sum to the given weight
        """
        data = castep.read_phonon_data(
            filename,
            read_eigenvectors=False,
            average_repeat_points=average_repeat_points)

        return cls.from_dict(data)

    @classmethod
    def from_phonopy(cls: Type[T], path: str = '.',
                     phonon_name: str = 'band.yaml',
                     phonon_format: Optional[str] = None,
                     summary_name: str = 'phonopy.yaml') -> T:
        """
        Reads precalculated phonon mode data from a Phonopy
        mesh/band/qpoints.yaml/hdf5 file. May also read from
        phonopy.yaml for structure information.

        Parameters
        ----------
        path
            Path to directory containing the file(s)
        phonon_name
            Name of Phonopy file including the frequencies
        phonon_format
            Format of the phonon_name file if it isn't obvious from the
            phonon_name extension, one of {'yaml', 'hdf5'}
        summary_name
            Name of Phonopy summary file to read the crystal information
            from. Crystal information in the phonon_name file takes
            priority, but if it isn't present, crystal information is
            read from summary_name instead
        """
        data = phonopy.read_phonon_data(
            path=path, phonon_name=phonon_name, phonon_format=phonon_format,
            summary_name=summary_name, read_eigenvectors=False)
        return cls.from_dict(data)



from argparse import ArgumentParser
from math import ceil
from typing import List, Optional

import matplotlib.style
import numpy as np

from euphonic import ureg, ForceConstants
from euphonic.cli.utils import (_calc_modes_kwargs, _compose_style,
                                _get_cli_parser,
                                _get_debye_waller, _get_energy_bins,
                                _get_q_distance, _get_pdos_weighting,
                                _arrange_pdos_groups, _plot_label_kwargs)
from euphonic.cli.utils import (load_data_from_file, get_args,
                                matplotlib_save_or_show)
import euphonic.plot
from euphonic.powder import (sample_sphere_dos, sample_sphere_pdos,
                             sample_sphere_structure_factor)
from euphonic.spectra import apply_kinematic_constraints
from euphonic.styles import base_style, intensity_widget_style
import euphonic.util

# Dummy tqdm function if tqdm progress bars unavailable
try:
    from tqdm.auto import tqdm
except ModuleNotFoundError:
    try:
        from tqdm import tqdm
    except ModuleNotFoundError:
        def tqdm(sequence):
            return sequence


def get_parser() -> ArgumentParser:

    parser, sections = _get_cli_parser(
        features={'read-fc', 'pdos-weighting', 'ins-weighting',
                  'powder', 'plotting', 'ebins', 'q-e', 'map',
                  'kinematic'})

    sections['q'].description = (
        '"GRID" options relate to Monkhorst-Pack sampling for the '
        'Debye-Waller factor, and only apply when --weighting=coherent and '
        '--temperature is set. "NPTS" options determine spherical groups of '
        'q-points for powder-averaging. "Q" options relate to the sphere '
        'sizes (i.e. radial distances).')

    sections['q'].add_argument('--q-min', type=float, default=0., dest='q_min',
                               help="Minimum |q| in 1/LENGTH_UNIT")
    sections['q'].add_argument('--q-max', type=float, default=3., dest='q_max',
                               help="Maximum |q| in 1/LENGTH_UNIT")
    sections['plotting'].add_argument('--no-widgets', action='store_true',
                                      dest='disable_widgets', default=False,
                                      help=("Don't use Matplotlib widgets to "
                                            "enable interactive setting of "
                                            "colormap intensity limits"))


    kinematic = parser.add_argument_group('Kinematic constraints')
    e_lims = kinematic.add_mutually_exclusive_group()

    e_lims.add_argument('--e-incident', '--e-i', dest='e_i',
                        type=float, default=None,
                        help="Incident energy for direct-geometry constraints")
    e_lims.add_argument('--e-final', '--e-f', dest='e_f',
                        type=float, default=None,
                        help="Final energy for indirect-geometry constraints")
    kinematic.add_argument(
        '--angle-range', nargs=2, type=float, dest='angle_range',
        default=[0., 180.],
        help=("Range of scattering angles (2θ) in degrees. These lower/upper "
              "bounds are used with incident/final energy to determine "
              "accessible (|q|, ω) region."))
    return parser


def main(params: Optional[List[str]] = None) -> None:
    args = get_args(get_parser(), params)
    calc_modes_kwargs = _calc_modes_kwargs(args)

    # Make sure we get an error if accessing NPTS inappropriately
    if args.npts_density is not None:
        args.npts = None

    fc = load_data_from_file(args.filename, verbose=True)
    if not isinstance(fc, ForceConstants):
        raise TypeError('Force constants are required to use the '
                        'euphonic-powder-map tool')
    if args.pdos is not None and args.weighting == 'coherent':
        raise ValueError('"--pdos" is only compatible with '
                         '"--weighting" options that include dos')
    print("Setting up dimensions...")

    q_min = _get_q_distance(args.length_unit, args.q_min)
    q_max = _get_q_distance(args.length_unit, args.q_max)
    recip_length_unit = q_min.units

    n_q_bins = ceil((args.q_max - args.q_min) / args.q_spacing)
    q_bin_edges = np.linspace(q_min.magnitude, q_max.magnitude, n_q_bins + 1,
                              endpoint=True) * recip_length_unit
    q_bin_centers = (q_bin_edges[:-1] + q_bin_edges[1:]) / 2

    # Use X-point modes to estimate frequency range, set up energy bins
    # (Not Gamma in case there are only 3 branches; value would be zero!)
    modes = fc.calculate_qpoint_frequencies(
        np.array([[0., 0., 0.5]]), **calc_modes_kwargs)
    modes.frequencies_unit = args.energy_unit

    if args.e_i is not None and args.e_max is None:
        emax = args.e_i
    else:
        emax = args.e_max

    energy_bins = _get_energy_bins(
        modes, args.ebins + 1, emin=args.e_min, emax=emax,
        headroom=1.2)  # Generous headroom as we only checked one q-point

    if args.weighting in ('coherent',):
        # Compute Debye-Waller factor once for re-use at each mod(q)
        # (If temperature is not set, this will be None.)
        if args.temperature is not None:
            temperature = args.temperature * ureg('K')
            dw = _get_debye_waller(temperature, fc, grid=args.grid,
                                   grid_spacing=(args.grid_spacing
                                                 * recip_length_unit),
                                   **calc_modes_kwargs)
        else:
            temperature = None
            dw = None

    print(f"Sampling {n_q_bins} |q| shells between {q_min:~P} and {q_max:~P}")

    z_data = np.empty((n_q_bins, len(energy_bins) - 1))

    for q_index in tqdm(range(n_q_bins)):
        q = q_bin_centers[q_index]

        if args.npts_density is not None:
            npts = ceil(args.npts_density * (q / recip_length_unit)**2)
            npts = max(args.npts_min,
                       min(args.npts_max, npts))
        else:
            npts = args.npts

        if args.weighting == 'dos' and args.pdos is None:
            spectrum_1d = sample_sphere_dos(
                fc, q,
                npts=npts, sampling=args.sampling, jitter=args.jitter,
                energy_bins=energy_bins,
                **calc_modes_kwargs)
        elif 'dos' in args.weighting:
            spectrum_1d_col = sample_sphere_pdos(
                    fc, q,
                    npts=npts, sampling=args.sampling, jitter=args.jitter,
                    energy_bins=energy_bins,
                    weighting=_get_pdos_weighting(args.weighting),
                    **calc_modes_kwargs)
            spectrum_1d = _arrange_pdos_groups(spectrum_1d_col, args.pdos)
        elif args.weighting == 'coherent':
            spectrum_1d = sample_sphere_structure_factor(
                fc, q,
                dw=dw,
                temperature=temperature,
                sampling=args.sampling, jitter=args.jitter,
                npts=npts,
                energy_bins=energy_bins,
                **calc_modes_kwargs)

        z_data[q_index, :] = spectrum_1d.y_data.magnitude

    print(f"Final npts: {npts}")

    spectrum = euphonic.Spectrum2D(q_bin_edges, energy_bins,
                                   z_data * spectrum_1d.y_data.units)

    if args.q_broadening or args.energy_broadening:
        spectrum = spectrum.broaden(
            x_width=(args.q_broadening * recip_length_unit
                     if args.q_broadening else None),
            y_width=(args.energy_broadening * energy_bins.units
                     if args.energy_broadening else None),
            shape=args.shape)

    if not (args.e_i is None and args.e_f is None):
        print("Applying kinematic constraints")
        energy_unit = args.energy_unit
        e_i = args.e_i * ureg(energy_unit) if (args.e_i is not None) else None
        e_f = args.e_f * ureg(energy_unit) if (args.e_f is not None) else None
        spectrum = apply_kinematic_constraints(spectrum,
            e_i=e_i, e_f=e_f, angle_range=args.angle_range)

    print(f"Plotting figure: max intensity "
          f"{np.nanmax(spectrum.z_data.magnitude) * spectrum.z_data.units:~P}")
    plot_label_kwargs = _plot_label_kwargs(
        args, default_xlabel=f"|q| / {q_min.units:~P}",
        default_ylabel=f"Energy / {spectrum.y_data.units:~P}")

    if args.save_json:
        spectrum.to_json_file(args.save_json)
    if args.disable_widgets:
        base = [base_style]
    else:
        base = [base_style, intensity_widget_style]
    style = _compose_style(user_args=args, base=base)
    with matplotlib.style.context(style):
        fig = euphonic.plot.plot_2d(spectrum,
                                    vmin=args.vmin,
                                    vmax=args.vmax,
                                    **plot_label_kwargs)

        if args.disable_widgets is False:
            # TextBox only available from mpl 2.1.0
            try:
                from matplotlib.widgets import TextBox
            except ImportError:
                args.disable_widgets = True

        if args.disable_widgets is False:
            min_label = f'Min Intensity ({spectrum.z_data.units:~P})'
            max_label = f'Max Intensity ({spectrum.z_data.units:~P})'
            boxw = 0.15
            boxh = 0.05
            x0 = 0.1 + len(min_label)*0.01
            y0 = 0.025
            axmin = fig.add_axes([x0, y0, boxw, boxh])
            axmax = fig.add_axes([x0, y0 + 0.075, boxw, boxh])
            image = fig.get_axes()[0].images[0]
            cmin, cmax = image.get_clim()
            pad = 0.05
            fmt_str = '.2e' if cmax < 0.1 else '.2f'
            minbox = TextBox(axmin, min_label,
                             initial=f'{cmin:{fmt_str}}', label_pad=pad)
            maxbox = TextBox(axmax, max_label,
                             initial=f'{cmax:{fmt_str}}', label_pad=pad)

            def update_min(min_val):
                image.set_clim(vmin=float(min_val))
                fig.canvas.draw()

            def update_max(max_val):
                image.set_clim(vmax=float(max_val))
                fig.canvas.draw()
            minbox.on_submit(update_min)
            maxbox.on_submit(update_max)

        matplotlib_save_or_show(save_filename=args.save_to)



from argparse import ArgumentParser
from typing import List, Optional

import matplotlib.style
import numpy as np

import euphonic
from euphonic import ureg, Spectrum2D, QpointFrequencies, ForceConstants
import euphonic.plot
from euphonic.util import get_qpoint_labels
from euphonic.styles import base_style
from .utils import (_bands_from_force_constants, _calc_modes_kwargs,
                    _compose_style, _plot_label_kwargs,
                    get_args, _get_debye_waller,
                    _get_energy_bins, _get_q_distance,
                    _get_cli_parser, load_data_from_file,
                    matplotlib_save_or_show)


def main(params: Optional[List[str]] = None) -> None:
    args = get_args(get_parser(), params)
    calc_modes_kwargs = _calc_modes_kwargs(args)

    frequencies_only = (args.weighting != 'coherent')
    data = load_data_from_file(args.filename, verbose=True,
                               frequencies_only=frequencies_only)
    if not frequencies_only and type(data) is QpointFrequencies:
        raise TypeError('Eigenvectors are required to use '
                        '"--weighting coherent" option')
    if args.weighting.lower() == 'coherent' and args.temperature is not None:
        if not isinstance(data, ForceConstants):
            raise TypeError('Force constants data is required to generate '
                            'the Debye-Waller factor. Leave "--temperature" '
                            'unset if plotting precalculated phonon modes.')

    q_spacing = _get_q_distance(args.length_unit, args.q_spacing)
    recip_length_unit = q_spacing.units

    if isinstance(data, ForceConstants):
        print("Getting band path...")
        (modes, x_tick_labels, split_args) = _bands_from_force_constants(
            data, q_distance=q_spacing, insert_gamma=False,
            frequencies_only=frequencies_only,
            **calc_modes_kwargs)
    else:
        modes = data
        split_args = {'btol': args.btol}
        x_tick_labels = get_qpoint_labels(modes.qpts,
                                          cell=modes.crystal.to_spglib_cell())
    modes.frequencies_unit = args.energy_unit
    ebins = _get_energy_bins(modes, args.ebins + 1, emin=args.e_min,
                             emax=args.e_max)

    print("Computing intensities and generating 2D maps")

    if args.weighting.lower() == 'coherent':
        if args.temperature is not None:
            temperature = args.temperature * ureg('K')
            dw = _get_debye_waller(temperature, data,
                                   grid=args.grid,
                                   grid_spacing=(args.grid_spacing
                                                 * recip_length_unit),
                                   **calc_modes_kwargs)
        else:
            dw = None

        spectrum = (modes.calculate_structure_factor(dw=dw)
                    .calculate_sqw_map(ebins))

    elif args.weighting.lower() == 'dos':
        spectrum = modes.calculate_dos_map(ebins)

    if args.q_broadening or args.energy_broadening:
        spectrum = spectrum.broaden(
            x_width=(args.q_broadening * recip_length_unit
                     if args.q_broadening else None),
            y_width=(args.energy_broadening * ebins.units
                     if args.energy_broadening else None),
            shape=args.shape, method='convolve')

    print("Plotting figure")
    plot_label_kwargs = _plot_label_kwargs(
        args, default_ylabel=f"Energy / {spectrum.y_data.units:~P}")

    if x_tick_labels:
        spectrum.x_tick_labels = x_tick_labels

    spectra = spectrum.split(**split_args)  # type: List[Spectrum2D]
    if len(spectra) > 1:
        print(f"Found {len(spectra)} regions in q-point path")

    if args.save_json:
        spectrum.to_json_file(args.save_json)
    style = _compose_style(user_args=args, base=[base_style])
    with matplotlib.style.context(style):

        euphonic.plot.plot_2d(spectra,
                              vmin=args.vmin,
                              vmax=args.vmax,
                              **plot_label_kwargs)
        matplotlib_save_or_show(save_filename=args.save_to)


def get_parser() -> ArgumentParser:
    parser, sections = _get_cli_parser(
        features={'read-fc', 'read-modes', 'q-e', 'map', 'btol', 'ebins',
                  'ins-weighting', 'plotting'})
    parser.description = (
        'Plots a 2D intensity map from the file provided. If a force '
        'constants file is provided, a band structure path is '
        'generated using Seekpath')

    sections['q'].description = (
        '"GRID" options relate to Monkhorst-Pack sampling for the '
        'Debye-Waller factor, and only apply when --weighting=coherent '
        'and --temperature is set. "Q" options relate to the x-axis of '
        'spectrum data.')

    return parser



from typing import Optional, Sequence, Tuple, Union

try:
    import matplotlib.pyplot as plt
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from matplotlib.colors import Colormap, Normalize
    from matplotlib.image import NonUniformImage

except ModuleNotFoundError:
    raise ModuleNotFoundError(
        'Cannot import Matplotlib for plotting (maybe Matplotlib is '
        'not installed?). To install Euphonic\'s optional Matplotlib '
        'dependency, try:\n\npip install euphonic[matplotlib]\n')

import numpy as np

from euphonic import Quantity
from euphonic.spectra import Spectrum1D, Spectrum1DCollection, Spectrum2D

def plot_1d_to_axis(spectra: Union[Spectrum1D, Spectrum1DCollection],
                    ax: Axes, labels: Optional[Sequence[str]] = None,
                    **mplargs) -> None:
    """Plot a (collection of) 1D spectrum lines to matplotlib axis

    Where there are two identical x-values in a row, plotting will restart
    to avoid creating a vertical line

    Parameters
    ----------
    spectra
        Spectrum1D or Spectrum1DCollection to plot
    ax
        Matplotlib axes to which spectra will be drawn
    labels
        A sequence of labels corresponding to the sequence of lines in
        spectra, used to label each line. If this is None, the
        label(s) contained in spectra.metadata['label'] (Spectrum1D) or
        spectra.metadata['line_data'][i]['label']
        (Spectrum1DCollection) will be used. To disable labelling for a
        specific line, pass an empty string.
    **mplargs
        Keyword arguments passed to Axes.plot
    """

    if isinstance(spectra, Spectrum1D):
        return plot_1d_to_axis(Spectrum1DCollection.from_spectra([spectra]),
                               ax=ax, labels=labels, **mplargs)

    try:
        assert isinstance(spectra, Spectrum1DCollection)
    except AssertionError:
        raise TypeError("spectra should be a Spectrum1D or "
                        "Spectrum1DCollection")

    if isinstance(labels, str):
        labels = [labels]
    if labels is not None and len(labels) != len(spectra):
        raise ValueError(
            f"The length of labels (got {len(labels)}) should be the "
            f"same as the number of lines to plot (got {len(spectra)})")

    # Find where there are two identical x_data points in a row
    breakpoints = (np.where(spectra.x_data.magnitude[:-1]
                            == spectra.x_data.magnitude[1:])[0]
                   + 1).tolist()
    breakpoints = [0] + breakpoints + [None]

    if labels is None:
        labels = [spec.metadata.get('label', None) for spec in spectra]

    for label, spectrum in zip(labels, spectra):
        # Plot each line in segments
        for x0, x1 in zip(breakpoints[:-1], breakpoints[1:]):
            # Keep colour consistent across segments
            if x0 == 0:
                color = None
            else:
                # Only add legend label to the first segment
                label = None
                color = p[-1].get_color()

            p = ax.plot(spectrum.get_bin_centres().magnitude[x0:x1],
                        spectrum.y_data.magnitude[x0:x1],
                        color=color, label=label, **mplargs)

    # Update legend if it exists, in case new labels have been added
    legend = ax.get_legend()
    if legend is not None:
        ax.legend()

    ax.set_xlim(left=min(spectra.x_data.magnitude),
                right=max(spectra.x_data.magnitude))

    _set_x_tick_labels(ax, spectra.x_tick_labels, spectra.x_data)


def plot_1d(spectra: Union[Spectrum1D,
                           Spectrum1DCollection,
                           Sequence[Spectrum1D],
                           Sequence[Spectrum1DCollection]],
            title: str = '',
            xlabel: str = '',
            ylabel: str = '',
            ymin: Optional[float] = None,
            ymax: Optional[float] = None,
            labels: Optional[Sequence[str]] = None,
            **line_kwargs) -> Figure:
    """
    Creates a Matplotlib figure for a Spectrum1D object, or multiple
    Spectrum1D objects to be plotted on the same axes

    Parameters
    ----------
    spectra
        1D data to plot. Spectrum1D objects contain a single line, while
        Spectrum1DCollection is suitable for plotting multiple lines
        simultaneously (e.g. band structures).

        Data split across several regions should be provided as a sequence of
        spectrum objects::

            [Spectrum1D, Spectrum1D, ...]

        or::

            [Spectrum1DCollection, Spectrum1DCollection, ...]

        Where each segment will be plotted on a separate subplot. (This
        approach is mainly used to handle discontinuities in Brillouin-zone
        band structures, so the subplot widths will be based on the x-axis
        ranges.)

    title
        Plot title
    xlabel
        X-axis label
    ylabel
        Y-axis label
    ymin
        Minimum value on the y-axis. Can be useful to set y-axis minimum
        to 0 for energy, for example.
    ymax
        Maximum value on the y-axis.
    labels
        A sequence of labels corresponding to the sequence of lines in
        spectra, used to label each line. If this is None, the
        label(s) contained in spectra.metadata['label'] (Spectrum1D) or
        spectra.metadata['line_data'][i]['label']
        (Spectrum1DCollection) will be used. To disable labelling for a
        specific line, pass an empty string.
    **line_kwargs
        matplotlib.line.Line2D properties, optional
        Used in the axes.plot command to specify properties like
        linewidth, linestyle

    Returns
    -------
    fig : matplotlib.figure.Figure

    """
    if isinstance(spectra, (Spectrum1D, Spectrum1DCollection)):
        spectra = (spectra,)
    else:
        # Check units are consistent
        for spectrum in spectra[1:]:
            if spectrum.x_data_unit != spectra[0].x_data_unit:
                raise ValueError("Something went wrong: x data units are not "
                                 "consistent between spectrum subplots.")
            if spectrum.y_data_unit != spectra[0].y_data_unit:
                raise ValueError("Something went wrong: y data units are not "
                                 "consistent between spectrum subplots.")

    gridspec_kw = _get_gridspec_kw(spectra)
    fig, subplots = plt.subplots(1, len(spectra), sharey=True,
                                 gridspec_kw=gridspec_kw, squeeze=False)

    for i, (spectrum, ax) in enumerate(zip(spectra, subplots.flatten())):
        plot_1d_to_axis(spectrum, ax, labels, **line_kwargs)
        # To avoid an ugly empty legend, only use if there are labels to plot
        if i == 0:
            leg_handles, leg_labels = ax.get_legend_handles_labels()
            if len(leg_labels) > 0:
                ax.legend()
        ax.set_ylim(bottom=ymin, top=ymax)

    # Add an invisible large axis for common labels
    ax = fig.add_subplot(111, frameon=False)
    ax.grid(False)
    ax.tick_params(labelcolor="none", bottom=False, left=False)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    fig.suptitle(title)
    return fig


def plot_2d_to_axis(spectrum: Spectrum2D, ax: Axes,
                    cmap: Union[str, Colormap, None] = None,
                    interpolation: str = 'nearest',
                    norm: Optional[Normalize] = None,
                    ) -> NonUniformImage:
    """Plot Spectrum2D object to Axes

    Parameters
    ----------
    spectrum
        2D data object for plotting as NonUniformImage. The x_tick_labels
        attribute will be used to mark labelled points.
    ax
        Matplotlib axes to which image will be drawn
    cmap
        Matplotlib colormap or registered colormap name
    interpolation
        Interpolation method: 'nearest' or 'bilinear' for a pixellated or
        smooth result
    norm
        Matplotlib normalization object; set this in order to ensure separate
        plots are on the same colour scale.

    """
    x_unit = spectrum.x_data_unit
    y_unit = spectrum.y_data_unit
    z_unit = spectrum.z_data_unit

    x_bins = spectrum.get_bin_edges('x').to(x_unit).magnitude
    y_bins = spectrum.get_bin_edges('y').to(y_unit).magnitude

    image = NonUniformImage(ax, interpolation=interpolation,
                            extent=(min(x_bins), max(x_bins),
                                    min(y_bins), max(y_bins)),
                            cmap=cmap)
    if norm is not None:
        image.set_norm(norm)

    image.set_data(spectrum.get_bin_centres('x').to(x_unit).magnitude,
                   spectrum.get_bin_centres('y').to(y_unit).magnitude,
                   spectrum.z_data.to(z_unit).magnitude.T)
    ax.add_image(image)
    ax.set_xlim(min(x_bins), max(x_bins))
    ax.set_ylim(min(y_bins), max(y_bins))

    _set_x_tick_labels(ax, spectrum.x_tick_labels, spectrum.x_data)

    return image


def plot_2d(spectra: Union[Spectrum2D, Sequence[Spectrum2D]],
            vmin: Optional[float] = None,
            vmax: Optional[float] = None,
            cmap: Optional[Union[str, Colormap]] = None,
            title: str = '',
            xlabel: str = '',
            ylabel: str = '') -> Figure:
    """
    Creates a Matplotlib figure for a Spectrum2D object

    Parameters
    ----------
    spectra
        Containing the 2D data to plot. If a sequence of Spectrum2D is given,
        they will be plotted from right-to-left as separate subplots. This is
        recommended for band structure/dispersion plots with discontinuous
        regions.
    vmin
        Minimum of data range for colormap. See Matplotlib imshow docs
    vmax
        Maximum of data range for colormap. See Matplotlib imshow docs
    cmap
        Which colormap to use, see Matplotlib docs
    title
        Set a title for the Figure.
    xlabel
        X-axis label
    ylabel
        Y-axis label

    Returns
    -------
    fig : matplotlib.figure.Figure
        The Figure instance
    """
    # Wrap a bare spectrum in list so treatment is consistent with sequences
    if isinstance(spectra, Spectrum2D):
        spectra = [spectra]

    x_unit = spectra[0].x_data.units

    def _get_q_range(data: Quantity) -> float:
        dimensionless_data = data.to(x_unit).magnitude
        return dimensionless_data[-1] - dimensionless_data[0]
    widths = [_get_q_range(spectrum.x_data) for spectrum in spectra]

    fig, axes = plt.subplots(ncols=len(spectra), nrows=1,
                             gridspec_kw={'width_ratios': widths},
                             sharey=True, squeeze=False)

    intensity_unit = spectra[0].z_data.units

    def _get_minmax_intensity(spectrum: Spectrum2D) -> Tuple[float, float]:
        dimensionless_data = spectrum.z_data.to(intensity_unit).magnitude
        assert isinstance(dimensionless_data, np.ndarray)
        return np.nanmin(dimensionless_data), np.nanmax(dimensionless_data)
    min_z_list, max_z_list = zip(*map(_get_minmax_intensity, spectra))
    if vmin is None:
        vmin = min(min_z_list)
    if vmax is None:
        vmax = max(max_z_list)

    norm = Normalize(vmin=vmin, vmax=vmax)

    for spectrum, ax in zip(spectra, axes.flatten()):
        plot_2d_to_axis(spectrum, ax, cmap=cmap, norm=norm)

    # Add an invisible large axis for common labels
    ax = fig.add_subplot(111, frameon=False)
    ax.grid(False)
    ax.tick_params(labelcolor="none", bottom=False, left=False)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    fig.suptitle(title)

    return fig


def _set_x_tick_labels(ax: Axes,
                       x_tick_labels: Optional[Sequence[Tuple[int, str]]],
                       x_data: Quantity) -> None:
    if x_tick_labels is not None:
        locs, labels = [list(x) for x in zip(*x_tick_labels)]
        x_values = x_data.magnitude  # type: np.ndarray
        ax.set_xticks(x_values[locs])

        # Rotate long tick labels
        if len(max(labels, key=len)) >= 11:
            ax.set_xticklabels(labels, rotation=90)
        else:
            ax.set_xticklabels(labels)


def _get_gridspec_kw(spectra: Sequence[Union[Spectrum1D,
                                             Spectrum1DCollection]]):
    """
    Creates a dictionary of gridspec_kw to be passed to
    matplotlib.pyplot.subplots

    Parameters
    ----------
    spectra
        series of spectral data containers corresponding to subplots

    Returns
    -------
    gridspec_kw : dict
        Contains key 'width_ratios' which is a list of subplot widths.
        Required so the x-scale is the same for each subplot
    """
    # Get width ratios so that the x-scale is the same for each subplot
    subplot_widths = [max(spectrum.x_data.magnitude)
                      - min(spectrum.x_data.magnitude)
                      for spectrum in spectra]
    gridspec_kw = dict(width_ratios=[w / subplot_widths[0]
                                     for w in subplot_widths])
    return gridspec_kw



import copy
import json
import os
from typing import Dict, Type, TypeVar, Sequence, Any

import numpy as np

from euphonic import ureg, Quantity, __version__

T = TypeVar('T')

def _to_json_dict(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert all keys in an output dictionary to a JSON serialisable
    format
    """
    for key, val in dictionary.items():
        if isinstance(val, np.ndarray):
            if val.dtype == np.complex128:
                val = val.view(np.float64).reshape(val.shape + (2,))
            dictionary[key] = val.tolist()
        elif isinstance(val, dict):
            if key != 'metadata':
                dictionary[key] = _to_json_dict(val)
    return dictionary


def _from_json_dict(dictionary: Dict[str, Any],
                    type_dict: Dict[str, Type[Any]] = {}) -> Dict[str, Any]:
    """
    For a dictionary read from a JSON file, convert all list key values
    to that specified in type_dict. If not specified in type_dict, just
    uses the default conversion provided by np.array(list)
    """
    for key, val in dictionary.items():
        if isinstance(val, list):
            if key in type_dict and type_dict[key] == tuple:
                dictionary[key] = [tuple(x) for x in val]
            elif key in type_dict and type_dict[key] == np.complex128:
                dictionary[key] = np.array(
                    val, dtype=np.float64).view(np.complex128).squeeze()
            else:
                dictionary[key] = np.array(val)
        elif isinstance(val, dict):
            if key != 'metadata':
                dictionary[key] = _from_json_dict(
                    val, type_dict)
    return dictionary


def _obj_to_dict(obj: T, attrs: Sequence[str]) -> Dict[str, Any]:
    dout = {}
    for attr in attrs:
        val = getattr(obj, attr)
        # Don't write None values to dict
        if val is None:
            continue
        # Don't write empty metadata to dict
        if attr == 'metadata' and not val:
            continue

        if isinstance(val, np.ndarray):
            dout[attr] = np.copy(val)
        elif isinstance(val, list):
            dout[attr] = val.copy()
        elif isinstance(val, Quantity):
            if isinstance(val.magnitude, np.ndarray):
                dout[attr] = np.copy(val.magnitude)
            else:
                dout[attr] = val.magnitude
            dout[attr + '_unit'] = str(val.units)
        elif hasattr(val, 'to_dict'):
            dout[attr] = val.to_dict()
        else:
            dout[attr] = val
    return dout


def _process_dict(dictionary: Dict[str, Any], quantities: Sequence[str] = [],
                  optional: Sequence[str] = []) -> Dict[str, Any]:
    """
    Process an input dictionary for creating objects. Convert keys in
    'quantities' to Quantity objects, and if any 'optional' keys are
    missing, set them to None
    """
    dictionary = copy.deepcopy(dictionary)
    for okey in optional:
        if not okey in dictionary.keys():
            dictionary[okey] = None

    for qkey in quantities:
        val = dictionary.pop(qkey)
        if val is not None:
            val = val*ureg(dictionary.pop(qkey + '_unit'))
        dictionary[qkey] = val
    return dictionary


def _obj_to_json_file(obj: T, filename: str) -> None:
    """
    Generic function for writing to a JSON file from a Euphonic object
    """
    dout = _to_json_dict(obj.to_dict())
    dout['__euphonic_class__'] = obj.__class__.__name__
    dout['__euphonic_version__'] = __version__
    with open(filename, 'w') as f:
        json.dump(dout, f, indent=4, sort_keys=True)
    print(f'Written to {os.path.realpath(f.name)}')


def _obj_from_json_file(cls: Type[T], filename: str,
                        type_dict: Dict[str, Type[Any]] = {}) -> T:
    """
    Generic function for reading from a JSON file to a Euphonic object
    """
    with open(filename, 'r') as f:
        obj_dict = json.loads(f.read())
    obj_dict = _from_json_dict(obj_dict, type_dict)
    return cls.from_dict(obj_dict)



"""Functions for sampling properties in high-dimensional space.

These are implemented in a generic way for application in euphonic.powder
"""

from itertools import product
from typing import Iterator, Tuple

import numpy as np


_golden_ratio = (1 + np.sqrt(5)) / 2

#  Keep in mind that at this stage the implementations are intended for testing
#  and method development; clarity is prioritised over numerical efficiency


def golden_square(npts: int, offset: bool = True, jitter: bool = False
                  ) -> Iterator[Tuple[float, float]]:
    """Yield a series of well-distributed points in 2-D unit square

    These are obtained by the golden ratio method

    (x_i, y_i) = (i / npts, (i / Phi) % 1)

    Parameters
    ----------
    npts
        Number of points to distribute

    offset
        Offset x-coordinates away from origin. (This is recommended for
        spherical sampling by Marques et al. (2013))

    jitter
        Randomly displace points by a distance of up to 1/(2 sqrt(N))

    Returns
    -------
    Iterator[Tuple[float, float]

        Sequence of (x, y) pairs in range 0-1
    """

    if offset:
        x_offset = 1 / (2 * npts)
    else:
        x_offset = 0

    for i in range(npts):
        if jitter:
            displacement = (np.random.random(2) - 0.5) / np.sqrt(npts)
            delta_x, delta_y = displacement
        else:
            delta_x, delta_y = (0, 0)

        yield ((i / npts + delta_x + x_offset) % 1,
               ((i / _golden_ratio) + delta_y) % 1)


def regular_square(n_rows: int, n_cols: int,
                   offset: bool = True, jitter: bool = False
                   ) -> Iterator[Tuple[float, float]]:
    """Yield a regular grid of (x, y) points in 2-D unit square

    Parameters
    ----------

    n_rows
        number of rows
    n_cols
        number of columns
    offset
        offset points to avoid origin
    jitter
        randomly displace each point within its "cell"

    Returns
    -------
    Iterator[tuple[float, float]]
        sequence of (x, y) pairs in range 0-1
    """

    x_spacing, y_spacing = (1 / n_cols), (1 / n_rows)
    x_sequence = np.arange(n_cols) * x_spacing
    y_sequence = np.arange(n_rows) * y_spacing

    if offset:
        x_offset, y_offset = (x_spacing / 2), (y_spacing / 2)
    else:
        x_offset, y_offset = 0, 0

    for x, y in product(x_sequence, y_sequence):
        if jitter:
            displacement = (np.random.random(2) - 0.5) * [x_spacing, y_spacing]
            delta_x, delta_y = displacement
        else:
            delta_x, delta_y = 0, 0

        yield (x + x_offset + delta_x,
               y + y_offset + delta_y)


def _spherical_polar_to_cartesian(phi, theta):
    return (np.cos(phi) * np.sin(theta),
            np.sin(phi) * np.sin(theta),
            np.cos(theta))


def _square_to_spherical_polar(x: float, y: float) -> Tuple[float, float]:
    """Map from Cartesian square to spherical polar (phi, theta)"""
    theta = np.arccos(2 * x - 1)
    phi = 2 * np.pi * y
    return (phi, theta)


def golden_sphere(npts: int, cartesian: bool = True, jitter: bool = False,
                  ) -> Iterator[Tuple[float, float, float]]:
    """Yield a series of 3D points on unit sphere surface

    These resemble spherical Fibonacci point sets, using the golden
    ratio to obtain an appropriate point set for any positive inter NPTS.
    The method is outlined for "Quasi-Monte Carlo" sampling by
    Marques et al. (2013) DOI:10.1111/cgf.12190

    Parameters
    ----------

    npts
        Number of points at sphere surface

    cartesian
        Yield points in Cartesian coordinates. If False, the 3-tuple
        points are given in spherical coordinates.

    jitter
        Randomly displace points about their positions on surface

    Returns
    -------
    Iterator[Tuple[float, float, float]]

        Sequence of (x, y, z) coordinates (if cartesian=True) or
        (r, phi, theta) spherical coordinates.
    """
    for x, y in golden_square(npts, jitter=jitter):
        phi, theta = _square_to_spherical_polar(x, y)

        if cartesian:
            yield _spherical_polar_to_cartesian(phi, theta)
        else:
            yield (1, phi, theta)


def sphere_from_square_grid(n_rows: int, n_cols: int,
                            cartesian: bool = True, jitter: bool = False
                            ) -> Iterator[Tuple[float, float, float]]:
    """Yield a series of 3D points on a unit sphere surface

    The points are projected from a uniform 2-D grid

    Parameters
    ----------
    n_rows
        Number of rows in the Cartesian generating grid
    n_cols
        Number of columns in the Cartesian generating grid
    cartesian
        Yield points in Cartesian coordinates. If False, instead yield points
        in spherical coordinates.
    jitter
        Randomly displace each point within its own "cell" of the grid

    Returns
    -------
    Iterator[Tuple[float, float, float]]

        Sequence of (x, y, z) coordinates (if cartesian=True) or
        (r, phi, theta) spherical coordinates.
    """
    for x, y in regular_square(n_rows, n_cols, jitter=jitter):
        phi, theta = _square_to_spherical_polar(x, y)

        if cartesian:
            yield _spherical_polar_to_cartesian(phi, theta)
        else:
            yield (1, phi, theta)


def spherical_polar_grid(n_phi: int, n_theta: int,
                         cartesian: bool = True, jitter: bool = False,
                         ) -> Iterator[Tuple[float, float, float]]:
    """Yield a series of 3D points on a unit sphere surface

    The points form a grid in spherical coordinates

    Parameters
    ----------

    n_phi
        Number of longitude-like radial subdivisions at poles

    n_theta
        Number of lattitude-like rings of points dividing polar axis

    cartesian
        Yield points in Cartesian coordinates. If False, instead yield points
        in spherical coordinates.

    jitter
        Randomly displace each point within its own "cell" of the grid

    Returns
    -------
    Iterator[Tuple[float, float, float]]

        Sequence of (x, y, z) coordinates (if cartesian=True) or
        (r, phi, theta) spherical coordinates.
    """
    phi_sequence = np.linspace(-np.pi, np.pi, n_phi + 1)[:-1]
    phi_spacing = phi_sequence[1] - phi_sequence[0]

    # Offset theta samples to avoid redundancy at pole
    theta_sequence = (np.linspace(0, np.pi, n_theta + 1)[:-1]
                      + np.pi / (2 * n_theta))
    theta_spacing = theta_sequence[1] - theta_sequence[0]

    for phi, theta in product(phi_sequence, theta_sequence):
        if jitter:
            displacement = ((np.random.random(2) - 0.5)
                            * [phi_spacing, theta_spacing])
            phi, theta = [phi, theta] + displacement

        if cartesian:
            yield _spherical_polar_to_cartesian(phi, theta)
        else:
            yield (1, phi, theta)


def spherical_polar_improved(npts: int,
                             cartesian: bool = True, jitter: bool = False,
                             ) -> Iterator[Tuple[float, float, float]]:
    """Yield a series of 3D points on a unit sphere surface

    The points form rings of common theta in polar coordinates. However,
    the number of samples in each ring is scaled to achieve as uniform a
    sampling density as possible.

    Analytically we find that for evenly-divided theta, the average
    length of a "latitude" line is 4.

    To obtain parity between nearest-neighbours (NN) along each axis, we
    solve (NN distance along theta) = (NN distance along phi):

    4 n_theta / npts = pi / n_theta

    Exact solutions with integer npts will be an irrational number, so
    n_theta is rounded down. The requested npts is then distributed
    between the constant-theta rings, according to their circumference.

    Parameters
    ----------

    npts
        Number of points at sphere surface

    cartesian
        Yield points in Cartesian coordinates. If False, instead yield
        points in spherical coordinates.

    jitter 
        Randomly displace each point within its own "cell" of the
        irregular grid

    Returns
    -------
    Iterator[Tuple[float, float, float]]

        Sequence of (x, y, z) coordinates (if cartesian=True) or
        (r, phi, theta) spherical coordinates.

    Raises
    ------
    ValueError
        If the number of points is not supported by this method
        """
    if npts < 6:
        raise ValueError("This sampling scheme has a minimum of 6 points")

    # round from the solution of
    # for neighbouring points
    n_theta = int(np.sqrt(np.pi / 4 * npts))

    # Offset theta samples to avoid redundancy at pole
    theta_sequence = (np.linspace(0, np.pi, n_theta + 1)[:-1]
                      + np.pi / (2 * n_theta))
    theta_spacing = theta_sequence[1] - theta_sequence[0]

    # get the lengths of all theta bins and drop points in
    theta_circumferences = 2 * np.pi * np.abs(np.sin(theta_sequence))
    bin_edges = np.concatenate([[0], np.cumsum(theta_circumferences)])

    counts, _ = np.histogram(np.linspace(0, theta_circumferences.sum(), npts),
                             bins=bin_edges, density=False)

    for row_theta, count in zip(theta_sequence, counts):
        phi_sequence = np.linspace(-np.pi, np.pi, count + 1)[:-1]
        phi_spacing = phi_sequence[1] - phi_sequence[0]
        for phi in phi_sequence:
            if jitter:
                displacement = (np.random.random(2) - 0.5) * [phi_spacing,
                                                              theta_spacing]
                phi, theta = [phi, row_theta] + displacement
            else:
                theta = row_theta

            if cartesian:
                yield _spherical_polar_to_cartesian(phi, theta)
            else:
                yield (1, phi, theta)


def random_sphere(npts, cartesian: bool = True
                  ) -> Iterator[Tuple[float, float, float]]:
    """Yield a series of 3D points on a unit sphere surface

    Points are distributed randomly in polar coordinates: phi is
    evenly-distributed from 0 to 2pi while theta is scaled with an
    arccos function from an even distribution over the range (-1, 1), to
    account for warping at the poles.

    Parameters
    ----------

    npts
        Number of points at sphere surface

    cartesian
        Yield points in Cartesian coordinates. If False, instead yield
        points in spherical coordinates.

    Returns
    -------
    Iterator[Tuple[float, float, float]]

        Sequence of (x, y, z) coordinates (if cartesian=True) or
        (r, phi, theta) spherical coordinates.
        """

    points = np.random.random((npts, 2))

    for x, y in points:
        phi, theta = _square_to_spherical_polar(x, y)

        if cartesian:
            yield _spherical_polar_to_cartesian(phi, theta)
        else:
            yield (1, phi, theta)



from argparse import ArgumentParser
from typing import List, Optional

import matplotlib.style

from euphonic.plot import plot_1d
from euphonic.styles import base_style
from euphonic import Spectrum1D, ForceConstants, QpointFrequencies
from .utils import (load_data_from_file, get_args, _bands_from_force_constants,
                    _compose_style,
                    _get_q_distance, matplotlib_save_or_show, _get_cli_parser,
                    _calc_modes_kwargs, _plot_label_kwargs)


def main(params: Optional[List[str]] = None) -> None:
    args = get_args(get_parser(), params)

    frequencies_only = not args.reorder  # Need eigenvectors to reorder
    data = load_data_from_file(args.filename, verbose=True,
                               frequencies_only=frequencies_only)
    if not frequencies_only and type(data) is QpointFrequencies:
        raise TypeError(
            'Eigenvectors are required to use "--reorder" option')

    if isinstance(data, ForceConstants):
        print("Getting band path...")
        q_distance = _get_q_distance(args.length_unit, args.q_spacing)
        (bands, x_tick_labels, split_args) = _bands_from_force_constants(
            data, q_distance=q_distance, frequencies_only=frequencies_only,
            **_calc_modes_kwargs(args))
    else:
        bands = data
        split_args = {'btol': args.btol}
        x_tick_labels = None

    bands.frequencies_unit = args.energy_unit

    print("Mapping modes to 1D band-structure")
    if args.reorder:
        bands.reorder_frequencies()

    spectrum = bands.get_dispersion()

    plot_label_kwargs = _plot_label_kwargs(
        args, default_ylabel=f"Energy / {spectrum.y_data.units:~P}")

    if x_tick_labels:
        spectrum.x_tick_labels = x_tick_labels

    spectra = spectrum.split(**split_args)  # type: List[Spectrum1D]

    style = _compose_style(user_args=args,
                           base=[base_style])

    if args.save_json:
        spectrum.to_json_file(args.save_json)
    with matplotlib.style.context(style):
        _ = plot_1d(spectra,
                    ymin=args.e_min,
                    ymax=args.e_max,
                    **plot_label_kwargs)
        matplotlib_save_or_show(save_filename=args.save_to)


def get_parser() -> ArgumentParser:
    parser, _ = _get_cli_parser(features={'read-fc', 'read-modes', 'plotting',
                                          'q-e', 'btol'})
    parser.description = (
        'Plots a band structure from the file provided. If a force '
        'constants file is provided, a band structure path is '
        'generated using Seekpath')
    bands_group = parser.add_argument_group(
        'Band arguments',
        'Options related to plotting 1D bands ("spaghetti plots").')
    bands_group.add_argument(
        '--reorder',
        action='store_true',
        help=('Try to determine branch crossings from eigenvectors and'
              ' rearrange frequencies accordingly'))
    return parser



import os
import shutil
import warnings

import versioneer
from setuptools import setup, Extension, Distribution
from setuptools.command.install import install


# As the C extension is optional, this is not detected when building
# wheels and they are incorrectly marked as universal. Explicitly
# specify the distribution is not pure Python
try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False
except ImportError:
    bdist_wheel = None


class InstallCommand(install):
    user_options = install.user_options + [('python-only', None, 'Install only Python')]

    def initialize_options(self):
        install.initialize_options(self)
        self.python_only = 0

    def finalize_options(self):
        install.finalize_options(self)
        if not bool(self.python_only):
            self.distribution.ext_modules = [get_c_extension()]
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


def get_c_extension():
    import os
    import numpy as np
    from sys import platform
    import subprocess
    include_dirs = [np.get_include(), 'c']
    sources = ['c/_euphonic.c', 'c/dyn_mat.c', 'c/util.c', 'c/py_util.c',
               'c/load_libs.c']
    if platform == 'win32':
        # Windows - assume MSVC compiler
        compile_args = ['/openmp']
        link_args = None
    elif platform == 'darwin':
        # OSX - if CC not set, assume brew install llvm
        try:
            brew_prefix_cmd_return = subprocess.run(["brew", "--prefix"],
                                                    stdout=subprocess.PIPE)
            brew_prefix = brew_prefix_cmd_return.stdout.decode("utf-8").strip()
        except FileNotFoundError:
            brew_prefix = None

        if brew_prefix and not os.environ.get('CC'):
            os.environ['CC'] = '{}/opt/llvm/bin/clang'.format(brew_prefix)
            link_args = ['-L{}/opt/llvm/lib'.format(brew_prefix), '-fopenmp']
        else:
            link_args = ['-fopenmp']

        compile_args = ['-fopenmp']            

    else:
        # Linux - assume gcc if CC not set
        if not os.environ.get('CC'):
            os.environ['CC'] = 'gcc'

        compile_args = ['-fopenmp']
        link_args = ['-fopenmp']

    euphonic_c_extension = Extension(
        'euphonic._euphonic',
        extra_compile_args=compile_args,
        extra_link_args=link_args,
        include_dirs=include_dirs,
        sources=sources
    )
    return euphonic_c_extension


def run_setup():

    license_file = 'LICENSE'
    # A list of files outside the Euphonic package directory that should
    # be included in the installation in site-packages. They must
    # also be added to MANIFEST.in
    ex_install_files = [license_file,
                        'CITATION.cff']
    # MANIFEST.in will add any included files to the site-packages
    # installation, but only if they are inside the package directory,
    # so temporarily copy them there if needed
    for ex_install_file in ex_install_files:
        try:
            shutil.copy(ex_install_file, 'euphonic')
        except (PermissionError, OSError) as err:
            warnings.warn(f'{err}', stacklevel=2)

    packages = ['euphonic',
                'euphonic.cli',
                'euphonic.readers',
                'euphonic.data',
                'euphonic.styles']

    with open('README.rst', 'r') as f:
        long_description = f.read()


    cmdclass = versioneer.get_cmdclass()
    cmdclass['install'] = InstallCommand
    cmdclass['bdist_wheel'] = bdist_wheel

    setup(
        name='euphonic',
        version=versioneer.get_version(),
        cmdclass=cmdclass,
        author='Rebecca Fair',
        author_email='rebecca.fair@stfc.ac.uk',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Programming Language :: C',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering :: Physics',
            ],
        description=(
            'Euphonic calculates phonon bandstructures and inelastic '
            'neutron scattering intensities from modelling code output '
            '(e.g. CASTEP)'),
        license='GPLv3',
        license_files=(license_file,),
        long_description=long_description,
        long_description_content_type='text/x-rst',
        url='https://github.com/pace-neutrons/Euphonic',
        packages=packages,
        include_package_data=True,
        install_requires=[
            'numpy>=1.14.5',
            'scipy>=1.1.0',
            'seekpath>=1.1.0',
            'spglib>=1.9.4',
            'pint>=0.10.1',
            'importlib_resources>=1.3.0',
            'threadpoolctl>=1.0.0'
        ],
        extras_require={
            'matplotlib': ['matplotlib>=2.2.2'],
            'phonopy_reader': ['h5py>=2.8.0', 'PyYAML>=3.13']
        },
        entry_points={'console_scripts': [
            'euphonic-dispersion = euphonic.cli.dispersion:main',
            'euphonic-dos = euphonic.cli.dos:main',
            'euphonic-optimise-dipole-parameter = euphonic.cli.optimise_dipole_parameter:main',
            'euphonic-show-sampling = euphonic.cli.show_sampling:main',
            'euphonic-intensity-map = euphonic.cli.intensity_map:main',
            'euphonic-powder-map = euphonic.cli.powder_map:main']}
    )

    for ex_install_file in ex_install_files:
        # Only delete file if there is another copy in the current dir
        # (e.g. from a git clone) in a PyPI dist the only copy will be
        # in the euphonic subdir, we don't want to delete that!
        if os.path.isfile(ex_install_file):
            try:
                os.remove(os.path.join('euphonic', ex_install_file))
            except (PermissionError, OSError, FileNotFoundError) as err:
                warnings.warn(f'{err}', stacklevel=2)

run_setup()



from abc import ABC, abstractmethod
import collections
import copy
import itertools
import math
import json
from numbers import Integral
from typing import (Any, Dict, List, Optional, overload,
                    Sequence, Tuple, TypeVar, Union, Type)
import warnings

from pint import DimensionalityError
import numpy as np
from scipy.ndimage import correlate1d, gaussian_filter

from euphonic.validate import _check_constructor_inputs, _check_unit_conversion
from euphonic.io import (_obj_to_json_file, _obj_from_json_file,
                         _obj_to_dict, _process_dict)
from euphonic.readers.castep import read_phonon_dos_data
from euphonic.util import _get_unique_elems_and_idx
from euphonic import ureg, Quantity, __version__


class Spectrum(ABC):
    T = TypeVar('T', bound='Spectrum')

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['x_data_unit', 'y_data_unit'])
        super().__setattr__(name, value)

    def _set_data(self, data: Quantity, attr_name: str) -> None:
        setattr(self, f'_{attr_name}_data',
                np.array(data.magnitude, dtype=np.float64))
        setattr(self, f'_internal_{attr_name}_data_unit', str(data.units))
        setattr(self, f'{attr_name}_data_unit', str(data.units))

    @property
    def x_data(self) -> Quantity:
        return self._x_data*ureg(self._internal_x_data_unit).to(
            self.x_data_unit)

    @x_data.setter
    def x_data(self, value: Quantity) -> None:
        self.x_data_unit = str(value.units)
        self._x_data = value.to(self._internal_x_data_unit).magnitude

    @property
    def y_data(self) -> Quantity:
        return self._y_data*ureg(self._internal_y_data_unit).to(
            self.y_data_unit)

    @y_data.setter
    def y_data(self, value: Quantity) -> None:
        self.y_data_unit = str(value.units)
        self._y_data = value.to(self._internal_y_data_unit).magnitude

    @property
    def x_tick_labels(self) -> List[Tuple[int, str]]:
        return self._x_tick_labels

    @x_tick_labels.setter
    def x_tick_labels(self, value: Sequence[Tuple[int, str]]) -> None:
        err_msg = ('x_tick_labels should be of type '
                   'Sequence[Tuple[int, str]] e.g. '
                   '[(0, "label1"), (5, "label2")]')
        if value is not None:
            if isinstance(value, Sequence):
                for elem in value:
                    if not (isinstance(elem, tuple)
                            and len(elem) == 2
                            and isinstance(elem[0], Integral)
                            and isinstance(elem[1], str)):
                        raise TypeError(err_msg)
                # Ensure indices in x_tick_labels are plain ints as
                # np.int64/32 etc. are not JSON serializable
                value = [(int(idx), label) for idx, label in value]
            else:
                raise TypeError(err_msg)
        self._x_tick_labels = value

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Write to dict using euphonic.io._obj_to_dict"""
        ...

    @classmethod
    @abstractmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """Initialise a Spectrum object from dictionary"""
        ...

    def to_json_file(self, filename: str) -> None:
        """
        Write to a JSON file. JSON fields are equivalent to
        from_dict keys

        Parameters
        ----------
        filename : str
            Name of the JSON file to write to
        """
        _obj_to_json_file(self, filename)

    @classmethod
    def from_json_file(cls: Type[T], filename: str) -> T:
        """
        Read from a JSON file. See from_dict for required fields

        Parameters
        ----------
        filename : str
            The file to read from
        """
        type_dict = {'x_tick_labels': tuple}
        return _obj_from_json_file(cls, filename, type_dict)

    @abstractmethod
    def _split_by_indices(self: T, indices: Union[Sequence[int], np.ndarray]
                          ) -> List[T]:
        """Split data along x axis at given indices"""
        ...

    def _split_by_tol(self: T, btol: float = 10.0) -> List[T]:
        """Split data along x-axis at detected breakpoints"""
        diff = np.diff(self.x_data.magnitude)
        median = np.median(diff)
        breakpoints = np.where((diff / median) > btol)[0] + 1
        return self._split_by_indices(breakpoints)

    @staticmethod
    def _ranges_from_indices(indices: Union[Sequence[int], np.ndarray]
                             ) -> List[Tuple[int, Optional[int]]]:
        """Convert a series of breakpoints to a series of slice ranges"""
        if len(indices) == 0:
            ranges = [(0, None)]
        else:
            ranges = [(0, indices[0])]
            if len(indices) > 1:
                ranges = ranges + [(indices[i], indices[i+1])
                                   for i in range(len(indices) - 1)]
            ranges = ranges + [(indices[-1], None)]
        return ranges

    @staticmethod
    def _cut_x_ticks(x_tick_labels: Union[Sequence[Tuple[int, str]], None],
                     x0: int,
                     x1: Union[int, None]) -> Union[List[Tuple[int, str]],
                                                    None]:
        """Crop and shift x labels to new x range"""
        if x_tick_labels is None:
            return None
        else:
            return [(int(x - x0), label) for (x, label) in x_tick_labels
                    if (x >= x0) and ((x1 is None) or (x < x1))]

    def split(self: T, indices: Union[Sequence[int], np.ndarray] = None,
              btol: float = None) -> List[T]:
        """Split to multiple spectra

        Data may be split by index. Alternatively, x-axis data may be
        split automatically by searching for unusually large gaps in
        energy values.  These would usually correspond to disconnected
        special-points in a band-structure diagram.

        Parameters
        ----------
        indices
            positions in data of breakpoints
        btol
            parameter used to identify breakpoints. This is a ratio
            between the gap in values and the median gap between
            neighbouring x-values. If neither indices nor btol is
            specified, this is set to 10.0.

        Returns
        -------
        split_spectra
            Separated spectrum regions. If passed to the appropriate
            functions in euphonic.plot this would be interpreted as a
            series of subplots.
        """

        if indices is None:
            if btol is None:
                btol = 10.0
            return self._split_by_tol(btol=btol)
        else:
            if btol is not None:
                raise ValueError("Cannot set both indices and btol")
            return self._split_by_indices(indices)

    @staticmethod
    def _broaden_data(data: np.ndarray, bin_centres: Sequence[np.ndarray],
                      widths: Sequence[float], shape: str = 'gauss',
                      method: Optional[str] = None
                      ) -> np.ndarray:
        """
        Broaden data along each axis by widths, returning the broadened
        data. Data can be 1D or 2D, and  the length of bin_centres and
        widths should match the number of dimensions in data.
        bin_centres and widths are assumed to be in the same units
        """
        shape_opts = ('gauss', 'lorentz')
        if shape not in shape_opts:
            raise ValueError(f'Invalid value for shape, got {shape}, '
                             f'should be one of {shape_opts}')
        method_opts = ('convolve', None)
        if method not in method_opts:
            raise ValueError(f'Invalid value for method, got {method}, '
                             f'should be one of {method_opts}')

        # We only want to check for unequal bins if using a method that
        # is not correct for unequal bins (currently this is the only
        # option but leave the 'if' anyway for future implementations)
        if method in ('convolve', None):
            axes = ['x_data', 'y_data']
            unequal_bin_axes = []
            for ax, (width, bin_data) in enumerate(zip(widths, bin_centres)):
                if width is not None:
                    bin_widths = np.diff(bin_data)
                    if not np.all(np.isclose(bin_widths, bin_widths[0])):
                        unequal_bin_axes += [axes[ax]]
            if len(unequal_bin_axes) > 0:
                msg = (f'{" and ".join(unequal_bin_axes)} bin widths are '
                       f'not equal, so broadening by convolution may give '
                       f'incorrect results.')
                if method is None:
                    raise ValueError(
                        msg + ' If you still want to broaden by convolution '
                        'please explicitly use the method="convolve" option.')
                else:
                    warnings.warn(msg, stacklevel=3)

        if shape == 'gauss':
            sigmas = []
            for width, bin_data in zip(widths, bin_centres):
                if width is None:
                    sigmas.append(0.0)
                else:
                    sigmas.append(Spectrum._gfwhm_to_sigma(width, bin_data))
            data_broadened = gaussian_filter(data, sigmas, mode='constant')
        elif shape == 'lorentz':
            data_broadened = data
            for ax, (width, bin_data) in enumerate(zip(widths, bin_centres)):
                if width is not None:
                    broadening = _distribution_1d(bin_data, width)
                    data_broadened = correlate1d(data_broadened, broadening,
                                                 mode='constant', axis=ax)
        return data_broadened

    @staticmethod
    def _gfwhm_to_sigma(fwhm: float, ax_bin_centres: np.ndarray) -> float:
        """
        Convert a Gaussian FWHM to sigma in units of the mean ax bin size

        Parameters
        ----------
        fwhm
            The Gaussian broadening FWHM
        ax_bin_centres
            Shape (n_bins,) float np.ndarray.
            The bin centres along the axis the broadening is applied to

        Returns
        -------
        sigma
            Sigma in units of mean ax bin size
        """
        sigma = fwhm/(2*math.sqrt(2*math.log(2)))
        mean_bin_size = np.mean(np.diff(ax_bin_centres))
        sigma_bin = sigma/mean_bin_size
        return sigma_bin

    @staticmethod
    def _bin_edges_to_centres(bin_edges: np.ndarray) -> np.ndarray:
        return bin_edges[:-1] + 0.5*np.diff(bin_edges)

    @staticmethod
    def _bin_centres_to_edges(bin_centres: np.ndarray) -> np.ndarray:
        return np.concatenate((
            [bin_centres[0]],
            (bin_centres[1:] + bin_centres[:-1])/2,
            [bin_centres[-1]]))

    @staticmethod
    def _is_bin_edge(data_length, bin_length) -> bool:
        """Determine if axis data are bin edges or centres"""
        if bin_length == data_length + 1:
            return True
        elif bin_length == data_length:
            return False
        else:
            raise ValueError((
                f'Unexpected data axis length {data_length} '
                f'for bin axis length {bin_length}'))

    def get_bin_edges(self) -> Quantity:
        """
        Get x-axis bin edges. If the size of x_data is one element larger
        than y_data, x_data is assumed to contain bin edges, but if x_data
        is the same size, x_data is assumed to contain bin centres and
        a conversion is made. In the conversion, the bin edges will
        not go outside the existing data bounds so the first and last
        bins may be half-size. In addition, each bin edge is assumed
        to be halfway between each bin centre, which may not be an
        accurate assumption in the case of differently sized bins.
        """
        # Need to use -1 index for y_data so it also works for
        # Spectrum1DCollection which has y_data shape (n_spectra, bins)
        if self._is_bin_edge(self.y_data.shape[-1], self.x_data.shape[0]):
            return self.x_data
        else:
            return self._bin_centres_to_edges(
                self.x_data.magnitude)*self.x_data.units

    def get_bin_centres(self) -> Quantity:
        """
        Get x-axis bin centres. If the size of x_data is the same size
        as y_data, x_data is assumed to contain bin centres, but if x_data
        is one element larger, x_data is assumed to contain bin edges and
        a conversion is made. In this conversion, the bin centres are
        assumed to be in the middle of each bin, which may not be an
        accurate assumption in the case of differently sized bins.
        """
        # Need to use -1 index for y_data so it also works for
        # Spectrum1DCollection which has y_data shape (n_spectra, bins)
        if self._is_bin_edge(self.y_data.shape[-1], self.x_data.shape[0]):
            return self._bin_edges_to_centres(
                self.x_data.magnitude)*self.x_data.units
        else:
            return self.x_data


class Spectrum1D(Spectrum):
    """
    For storing generic 1D spectra e.g. density of states

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_x_data,) float Quantity. The plot data in y
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str]]. Contains metadata about the
        spectrum. Keys should be strings and values should be strings
        or integers
        There are some functional keys:
          - 'label' : str. This is used label lines on a 1D plot
    """
    T = TypeVar('T', bound='Spectrum1D')

    def __init__(self, x_data: Quantity, y_data: Quantity,
                 x_tick_labels: Optional[Sequence[Tuple[int, str]]] = None,
                 metadata: Optional[Dict[str, Union[int, str]]] = None
                 ) -> None:
        """
        Parameters
        ----------
        x_data
            Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The
            x_data points (if size == (n_x_data,)) or x_data bin edges
            (if size == (n_x_data + 1,))
        y_data
            Shape (n_x_data,) float Quantity. The plot data in y
        x_tick_labels
            Special tick labels e.g. for high-symmetry points. The int
            refers to the index in x_data the label should be applied to
        metadata
            Contains metadata about the spectrum. Keys should be
            strings and values should be strings or integers
            There are some functional keys:
              - 'label' : str. This is used label lines on a 1D plot
        """
        _check_constructor_inputs(
            [y_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1,), (), ()],
            ['y_data', 'x_tick_labels', 'metadata'])
        ny = len(y_data)
        _check_constructor_inputs(
            [x_data], [Quantity],
            [[(ny,), (ny+1,)]], ['x_data'])
        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels
        self.metadata = {} if metadata is None else metadata

    def __add__(self, other: 'Spectrum1D') -> 'Spectrum1D':
        """
        Sums the y_data of two Spectrum1D objects together,
        their x_data axes must be equal, and their y_data must
        have compatible units and the same number of y_data
        entries

        Any metadata key/value pairs that are common to both
        spectra are retained, any others are discarded
        """
        spec_col = Spectrum1DCollection.from_spectra([self, other])
        return spec_col.sum()

    def _split_by_indices(self: T,
                          indices: Union[Sequence[int], np.ndarray]
                          ) -> List[T]:
        """Split data along x-axis at given indices"""
        ranges = self._ranges_from_indices(indices)

        return [type(self)(self.x_data[x0:x1], self.y_data[x0:x1],
                           x_tick_labels=self._cut_x_ticks(self.x_tick_labels,
                                                           x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary. See Spectrum1D.from_dict for details on
        keys/values

        Returns
        -------
        dict
        """
        return _obj_to_dict(self, ['x_data', 'y_data', 'x_tick_labels',
                                   'metadata'])

    def to_text_file(self, filename: str,
                     fmt: Optional[Union[str, Sequence[str]]] = None) -> None:
        """
        Write to a text file. The header contains metadata and unit
        information, the first column contains x_data and the second
        column contains y_data. Note that text files written in this
        format cannot be read back in by Euphonic.

        Parameters
        ----------
        filename
            Name of the text file to write to
        fmt
            A format specifier or sequence of specifiers (one for each
            column), to be passed to numpy.savetxt
        """
        spec = Spectrum1DCollection.from_spectra([self])
        spec.to_text_file(filename, fmt)

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """
        Convert a dictionary to a Spectrum1D object

        Parameters
        ----------
        d
            A dictionary with the following keys/values:

            - 'x_data': (n_x_data,) or (n_x_data + 1,) float ndarray
            - 'x_data_unit': str
            - 'y_data': (n_x_data,) float ndarray
            - 'y_data_unit': str

            There are also the following optional keys:

            - 'x_tick_labels': list of (int, string) tuples
            - 'metadata': dict

        Returns
        -------
        spectrum
        """
        d = _process_dict(d, quantities=['x_data', 'y_data'],
                          optional=['x_tick_labels', 'metadata'])
        return cls(d['x_data'], d['y_data'], x_tick_labels=d['x_tick_labels'],
                   metadata=d['metadata'])

    @classmethod
    def from_castep_phonon_dos(cls: Type[T], filename: str,
                               element: Optional[str] = None) -> T:
        """
        Reads DOS from a CASTEP .phonon_dos file

        Parameters
        ----------
        filename
            The path and name of the .phonon_dos file to read
        element
            Which element's PDOS to read. If not supplied reads
            the total DOS.
        """
        data = read_phonon_dos_data(filename)
        metadata = {}
        if element is None:
            element = 'Total'
        else:
            metadata['species'] = element
        metadata['label'] = element
        return cls(data['dos_bins']*ureg(data['dos_bins_unit']),
                   data['dos'][element]*ureg(data['dos_unit']),
                   metadata=metadata)

    def broaden(self: T, x_width: Quantity, shape: str = 'gauss',
                method: Optional[str] = None) -> T:
        """
        Broaden y_data and return a new broadened spectrum object

        Parameters
        ----------
        x_width
            Scalar float Quantity. The broadening FWHM
        shape
            One of {'gauss', 'lorentz'}. The broadening shape
        method
            Can be None or 'convolve'. Currently the only broadening
            method available is convolution with a broadening kernel,
            but this may not produce correct results for unequal bin
            widths. To use convolution anyway, explicitly set
            method='convolve'

        Returns
        -------
        broadened_spectrum
            A new Spectrum1D object with broadened y_data

        Raises
        ------
        ValueError
            If shape is not one of the allowed strings
        ValueError
            If method is None and bins are not of equal size
        """
        y_broadened = self._broaden_data(
            self.y_data.magnitude,
            [self.get_bin_centres().magnitude],
            [x_width.to(self.x_data_unit).magnitude],
            shape=shape, method=method)
        return type(self)(
            np.copy(self.x_data.magnitude)*ureg(self.x_data_unit),
            y_broadened*ureg(self.y_data_unit),
            copy.copy((self.x_tick_labels)),
            copy.copy(self.metadata))


LineData = Sequence[Dict[str, Union[str, int]]]


class Spectrum1DCollection(collections.abc.Sequence, Spectrum):
    """A collection of Spectrum1D with common x_data and x_tick_labels

    Intended for convenient storage of band structures, projected DOS
    etc. This object can be indexed or iterated to obtain individual
    Spectrum1D.

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_entries, n_x_data) float Quantity. The plot data in y,
        in rows corresponding to separate 1D spectra
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str, LineData]] or None. Contains metadata
        about the spectra. Keys should be strings and values should be
        strings or integers.
        There are some functional keys:
          - 'line_data' : LineData
                          This is a Sequence[Dict[str, Union[int, str]],
                          it contains metadata for each spectrum in
                          the collection, and must be of length
                          n_entries
    """
    T = TypeVar('T', bound='Spectrum1DCollection')

    def __init__(
            self, x_data: Quantity, y_data: Quantity,
            x_tick_labels: Optional[Sequence[Tuple[int, str]]] = None,
            metadata: Optional[Dict[str, Union[str, int, LineData]]
            ] = None) -> None:
        """
        Parameters
        ----------
        x_data
            Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The
            x_data points (if size == (n_x_data,)) or x_data bin edges
            (if size == (n_x_data + 1,))
        y_data
            Shape (n_entries, n_x_data) float Quantity. The plot data
            in y, in rows corresponding to separate 1D spectra
        x_tick_labels
            Special tick labels e.g. for high-symmetry points. The int
            refers to the index in x_data the label should be applied to
        metadata
            Contains metadata about the spectra. Keys should be
            strings and values should be strings or integers.
            There are some functional keys:
              - 'line_data' : LineData
                              This is a Sequence[Dict[str, Union[int, str]],
                              it contains metadata for each spectrum
                              in the collection, and must be of length
                              n_entries
        """

        _check_constructor_inputs(
            [y_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1, -1), (), ()],
            ['y_data', 'x_tick_labels', 'metadata'])
        ny = len(y_data[0])
        _check_constructor_inputs(
            [x_data], [Quantity],
            [[(ny,), (ny+1,)]], ['x_data'])

        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels
        if metadata and 'line_data' in metadata.keys():
            if len(metadata['line_data']) != len(y_data):
                raise ValueError(
                    f'y_data contains {len(y_data)} spectra, but '
                    f'metadata["line_data"] contains '
                    f'{len(metadata["line_data"])} entries')
        self.metadata = {} if metadata is None else metadata

    def __add__(self: T, other: T) -> T:
        """
        Appends the y_data of 2 Spectrum1DCollection objects,
        creating a single Spectrum1DCollection that contains
        the spectra from both objects. The two objects must
        have equal x_data axes, and their y_data must
        have compatible units and the same number of y_data
        entries

        Any metadata key/value pairs that are common to both
        spectra are retained in the top level dictionary, any
        others are put in the individual 'line_data' entries
        """
        return type(self).from_spectra([*self, *other])

    def _split_by_indices(self,
                          indices: Union[Sequence[int], np.ndarray]
                          ) -> List[T]:
        """Split data along x-axis at given indices"""

        ranges = self._ranges_from_indices(indices)

        return [type(self)(self.x_data[x0:x1], self.y_data[:, x0:x1],
                           x_tick_labels=self._cut_x_ticks(self.x_tick_labels,
                                                           x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    def __len__(self):
        return self.y_data.magnitude.shape[0]

    @overload
    def __getitem__(self, item: int) -> Spectrum1D:
        ...

    @overload  # noqa: F811
    def __getitem__(self, item: slice) -> T:
        ...

    @overload  # noqa: F811
    def __getitem__(self, item: Union[Sequence[int], np.ndarray]) -> T:
        ...

    def __getitem__(self, item: Union[int, slice, Sequence[int], np.ndarray]):  # noqa: F811
        new_metadata = copy.deepcopy(self.metadata)
        line_metadata = new_metadata.pop('line_data',
                                         [{} for _ in self._y_data])
        if isinstance(item, Integral):
            new_metadata.update(line_metadata[item])
            return Spectrum1D(self.x_data,
                              self.y_data[item, :],
                              x_tick_labels=self.x_tick_labels,
                              metadata=new_metadata)

        if isinstance(item, slice):
            if (item.stop is not None) and (item.stop >= len(self)):
                raise IndexError(f'index "{item.stop}" out of range')
            new_metadata.update(self._combine_metadata(line_metadata[item]))
        else:
            try:
                item = list(item)
                if not all([isinstance(i, Integral) for i in item]):
                    raise TypeError
            except TypeError:
                raise TypeError(f'Index "{item}" should be an integer, slice '
                                f'or sequence of ints')
            new_metadata.update(self._combine_metadata(
                [line_metadata[i] for i in item]))
        return type(self)(self.x_data,
                          self.y_data[item, :],
                          x_tick_labels=self.x_tick_labels,
                          metadata=new_metadata)

    @classmethod
    def from_spectra(cls: Type[T], spectra: Sequence[Spectrum1D]) -> T:
        if len(spectra) < 1:
            raise IndexError("At least one spectrum is needed for collection")

        def _type_check(spectrum):
            if not isinstance(spectrum, Spectrum1D):
                raise TypeError(
                    "from_spectra() requires a sequence of Spectrum1D")

        _type_check(spectra[0])
        x_data = spectra[0].x_data
        x_tick_labels = spectra[0].x_tick_labels
        y_data_length = len(spectra[0].y_data.magnitude)
        y_data_magnitude = np.empty((len(spectra), y_data_length))
        y_data_magnitude[0, :] = spectra[0].y_data.magnitude
        y_data_units = spectra[0].y_data.units

        for i, spectrum in enumerate(spectra[1:]):
            _type_check(spectrum)
            assert spectrum.y_data.units == y_data_units
            assert np.allclose(spectrum.x_data.magnitude, x_data.magnitude)
            assert spectrum.x_data.units == x_data.units
            assert spectrum.x_tick_labels == x_tick_labels
            y_data_magnitude[i + 1, :] = spectrum.y_data.magnitude

        metadata = cls._combine_metadata([spec.metadata for spec in spectra])
        y_data = Quantity(y_data_magnitude, y_data_units)
        return cls(x_data, y_data, x_tick_labels=x_tick_labels,
                   metadata=metadata)

    @staticmethod
    def _combine_metadata(all_metadata: Sequence[Dict[str, Union[int, str]]]
                          ) -> Dict[str, Union[int, str, LineData]]:
        """
        From a sequence of metadata dictionaries, combines all common
        key/value pairs into the top level of a metadata dictionary,
        all unmatching key/value pairs are put into the 'line_data'
        key, which is a list of metadata dicts for each element in
        all_metadata
        """
        # This is for combining multiple separate spectrum metadata,
        # they shouldn't have line_data
        for metadata in all_metadata:
            assert not 'line_data' in metadata.keys()
        # Combine all common key/value pairs
        combined_metadata = dict(
            set(all_metadata[0].items()).intersection(
                *[metadata.items() for metadata in all_metadata[1:]]))
        # Put all other per-spectrum metadata in line_data
        line_data = []
        for i, metadata in enumerate(all_metadata):
            sdata = copy.copy(metadata)
            for key in combined_metadata.keys():
                sdata.pop(key)
            line_data.append(sdata)
        if any(line_data):
            combined_metadata['line_data'] = line_data
        return combined_metadata

    def _combine_line_metadata(self, indices: Optional[Sequence[int]] = None
                               ) -> Dict[str, Any]:
        """
        For a metadata dictionary, combines all common key/value
        pairs in 'line_data' and puts them in a top-level dictionary.
        If indices is supplied, only those indices in 'line_data' are
        combined. Unmatching key/value pairs are discarded
        """
        line_data = self.metadata.get('line_data', [{}]*len(self))
        if indices is not None:
            line_data = [line_data[idx] for idx in indices]
        combined_line_data = self._combine_metadata(line_data)
        combined_line_data.pop('line_data', None)
        return combined_line_data

    def _get_line_data_vals(self, *line_data_keys: str) -> np.ndarray:
        """
        Get value of the key(s) for each element in
        metadata['line_data']. Returns a 1D array of tuples, where each
        tuple contains the value(s) for each key in line_data_keys, for
        a single element in metadata['line_data']. This allows easy
        grouping/selecting by specific keys

        For example, if we have a Spectrum1DCollection with the following metadata:
            {'desc': 'Quartz', 'line_data': [
                {'inst': 'LET', 'sample': 0, 'index': 1},
                {'inst': 'MAPS', 'sample': 1, 'index': 2},
                {'inst': 'MARI', 'sample': 1, 'index': 1},
            ]}
        Then:
            _get_line_data_vals('inst', 'sample') = [('LET', 0),
                                                     ('MAPS', 1),
                                                     ('MARI', 1)]

        Raises a KeyError if 'line_data' or the key doesn't exist
        """
        line_data = self.metadata['line_data']
        line_data_vals = np.empty(len(line_data), dtype=object)
        for i, data in enumerate(line_data):
            line_data_vals[i] = tuple([data[key] for key in line_data_keys])
        return line_data_vals

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary consistent with from_dict()

        Returns
        -------
        dict
        """
        return _obj_to_dict(self, ['x_data', 'y_data', 'x_tick_labels',
                                   'metadata'])

    def to_text_file(self, filename: str,
                     fmt: Optional[Union[str, Sequence[str]]] = None) -> None:
        """
        Write to a text file. The header contains metadata and unit
        information, the first column is x_data and each subsequent
        column is a y_data spectrum. Note that text files written in
        this format cannot be read back in by Euphonic.

        Parameters
        ----------
        filename
            Name of the text file to write to
        fmt
            A format specifier or sequence of specifiers (one for each
            column), to be passed to numpy.savetxt
        """
        common_metadata = copy.deepcopy(self.metadata)
        line_data = common_metadata.pop('line_data',
                                        [{} for i in self.y_data])
        header = [f'Generated by Euphonic {__version__}',
                  f'x_data in ({self.x_data.units})',
                  f'y_data in ({self.y_data.units})',
                  f'Common metadata: {json.dumps(common_metadata)}',
                  f'Column 1: x_data']
        for i, line in enumerate(line_data):
            header += [f'Column {i + 2}: y_data[{i}] {json.dumps(line)}']
        out_data = np.hstack((self.get_bin_centres().magnitude[:, np.newaxis],
                              self.y_data.transpose().magnitude))
        kwargs = {'header': '\n'.join(header)}
        if fmt is not None:
            kwargs['fmt'] = fmt
        np.savetxt(filename, out_data, **kwargs)

    @classmethod
    def from_dict(cls: Type[T], d) -> T:
        """
        Convert a dictionary to a Spectrum1DCollection object

        Parameters
        ----------
        d : dict
            A dictionary with the following keys/values:

            - 'x_data': (n_x_data,) or (n_x_data + 1,) float ndarray
            - 'x_data_unit': str
            - 'y_data': (n_x_data,) float ndarray
            - 'y_data_unit': str

            There are also the following optional keys:

            - 'x_tick_labels': list of (int, string) tuples
            - 'metadata': dict

        Returns
        -------
        spectrum_collection
        """
        d = _process_dict(d, quantities=['x_data', 'y_data'],
                          optional=['x_tick_labels', 'metadata'])
        return cls(d['x_data'], d['y_data'], x_tick_labels=d['x_tick_labels'],
                   metadata=d['metadata'])

    @classmethod
    def from_castep_phonon_dos(cls: Type[T], filename: str) -> T:
        """
        Reads total DOS and per-element PDOS from a CASTEP
        .phonon_dos file

        Parameters
        ----------
        filename
            The path and name of the .phonon_dos file to read
        """
        data = read_phonon_dos_data(filename)
        n_spectra = len(data['dos'].keys())
        metadata = {'line_data': [{} for x in range(n_spectra)]}
        for i, (species, dos_data) in enumerate(data['dos'].items()):
            if i == 0:
                y_data = np.zeros((n_spectra, len(dos_data)))
            y_data[i] = dos_data
            if species != 'Total':
                metadata['line_data'][i]['species'] = species
            metadata['line_data'][i]['label'] = species
        return Spectrum1DCollection(
            data['dos_bins']*ureg(data['dos_bins_unit']),
            y_data*ureg(data['dos_unit']),
            metadata=metadata)

    def broaden(self, x_width: Quantity, shape: str = 'gauss',
                method: Optional[str] = None) -> T:
        """
        Individually broaden each line in y_data, returning a new
        Spectrum1DCollection

        Parameters
        ----------
        x_width
            Scalar float Quantity. The broadening FWHM
        shape
            One of {'gauss', 'lorentz'}. The broadening shape
        method
            Can be None or 'convolve'. Currently the only broadening
            method available is convolution with a broadening kernel,
            but this may not produce correct results for unequal bin
            widths. To use convolution anyway, explicitly set
            method='convolve'

        Returns
        -------
        broadened_spectrum
            A new Spectrum1DCollection object with broadened y_data

        Raises
        ------
        ValueError
            If shape is not one of the allowed strings
        ValueError
            If method is None and bins are not of equal size
        """
        y_broadened = np.zeros_like(self.y_data)
        x_centres = [self.get_bin_centres().magnitude]
        x_width_calc = [x_width.to(self.x_data_unit).magnitude]
        for i, yi in enumerate(self.y_data.magnitude):
            y_broadened[i] = self._broaden_data(
                yi, x_centres, x_width_calc, shape=shape,
                method=method)
        return Spectrum1DCollection(
            np.copy(self.x_data.magnitude)*ureg(self.x_data_unit),
            y_broadened*ureg(self.y_data_unit),
            copy.copy((self.x_tick_labels)),
            copy.deepcopy(self.metadata))

    def group_by(self, *line_data_keys: str) -> T:
        """
        Group and sum y_data for each spectrum according to the values
        mapped to the specified keys in metadata['line_data']

        Parameters
        ----------
        line_data_keys
            The key(s) to group by. If only one line_data_key is
            supplied, if the value mapped to a key is the same for
            multiple spectra, they are placed in the same group and
            summed. If multiple line_data_keys are supplied, the values
            must be the same for all specified keys for them to be
            placed in the same group

        Returns
        -------
        grouped_spectrum
            A new Spectrum1DCollection with one line for each group. Any
            metadata in 'line_data' not common across all spectra in a
            group will be discarded
        """
        grouping_dict = _get_unique_elems_and_idx(
            self._get_line_data_vals(*line_data_keys))

        new_y_data = np.zeros((len(grouping_dict), self._y_data.shape[-1]))
        group_metadata = copy.deepcopy(self.metadata)
        group_metadata['line_data'] = [{}]*len(grouping_dict)
        for i, idxs in enumerate(grouping_dict.values()):
            # Look for any common key/values in grouped metadata
            group_i_metadata = self._combine_line_metadata(idxs)
            group_metadata['line_data'][i] = group_i_metadata
            new_y_data[i] = np.sum(self._y_data[idxs], axis=0)
        new_y_data = new_y_data*ureg(self._internal_y_data_unit).to(
            self.y_data_unit)

        return Spectrum1DCollection(self.x_data, new_y_data,
                                    x_tick_labels=self.x_tick_labels,
                                    metadata=group_metadata)

    def sum(self) -> Spectrum1D:
        """
        Sum y_data over all spectra

        Returns
        -------
        summed_spectrum
            A Spectrum1D created from the summed y_data. Any metadata
            in 'line_data' not common across all spectra will be
            discarded
        """
        metadata = copy.deepcopy(self.metadata)
        metadata.pop('line_data', None)
        metadata.update(self._combine_line_metadata())
        summed_y_data = np.sum(self._y_data, axis=0)*ureg(
            self._internal_y_data_unit).to(self.y_data_unit)
        return Spectrum1D(self.x_data, summed_y_data,
                          x_tick_labels=self.x_tick_labels,
                          metadata=metadata)

    def select(self, **select_key_values: Union[
            str, int, Sequence[str], Sequence[int]]) -> T:
        """
        Select spectra by their keys and values in metadata['line_data']

        Parameters
        ----------
        **select_key_values
            Key-value/values pairs in metadata['line_data'] describing
            which spectra to extract. For example, to select all spectra
            where metadata['line_data']['species'] = 'Na' or 'Cl' use
            spectrum.select(species=['Na', 'Cl']). To select 'Na' and
            'Cl' spectra where weighting is also coherent, use
            spectrum.select(species=['Na', 'Cl'], weighting='coherent')

        Returns
        -------
        selected_spectra
           A Spectrum1DCollection containing the selected spectra

        Raises
        ------
        ValueError
            If no matching spectra are found
        """
        select_val_dict = _get_unique_elems_and_idx(
            self._get_line_data_vals(*select_key_values.keys()))
        for key, value in select_key_values.items():
            if isinstance(value, (int, str)):
                select_key_values[key] = [value]
        value_combinations = itertools.product(*select_key_values.values())
        select_idx = np.array([], dtype=np.int32)
        for value_combo in value_combinations:
            try:
                idx = select_val_dict[value_combo]
            # Don't require every combination to match e.g.
            # spec.select(sample=[0, 2], inst=['MAPS', 'MARI'])
            # we don't want to error simply because there are no
            # inst='MAPS' and sample=2 combinations
            except KeyError:
                continue
            select_idx = np.append(select_idx, idx)
        if len(select_idx) == 0:
            raise ValueError(f'No spectra found with matching metadata '
                             f'for {select_key_values}')
        return self[select_idx]


class Spectrum2D(Spectrum):
    """
    For storing generic 2D spectra e.g. S(Q,w)

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_y_data,) or (n_y_data + 1,) float Quantity. The y_data
        bin points (if size == (n_y_data,)) or y_data bin edges (if size
        == (n_y_data + 1,))
    z_data
        Shape (n_x_data, n_y_data) float Quantity. The plot data in z
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str]]. Contains metadata about the
        spectrum. Keys should be strings and values should be strings
        or integers
    """
    T = TypeVar('T', bound='Spectrum2D')

    def __init__(self, x_data: Quantity, y_data: Quantity,
                 z_data: Quantity,
                 x_tick_labels: Optional[Sequence[Tuple[int, str]]] = None,
                 metadata: Optional[Dict[str, Union[int, str]]] = None
                 ) -> None:
        """
        Parameters
        ----------
        x_data
            Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The
            x_data points (if size == (n_x_data,)) or x_data bin edges
            (if size == (n_x_data + 1,))
        y_data
            Shape (n_y_data,) or (n_y_data + 1,) float Quantity. The
            y_data bin points (if size == (n_y_data,)) or y_data bin
            edges (if size == (n_y_data + 1,))
        z_data
            Shape (n_x_data, n_y_data) float Quantity. The plot data in z
        x_tick_labels
            Special tick labels e.g. for high-symmetry points. The int
            refers to the index in x_data the label should be applied to
        metadata
            Contains metadata about the spectrum. Keys should be
            strings and values should be strings or integers.
        """
        _check_constructor_inputs(
            [z_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1, -1), (), ()],
            ['z_data', 'x_tick_labels', 'metadata'])
        nx = z_data.shape[0]
        ny = z_data.shape[1]
        _check_constructor_inputs(
            [x_data, y_data],
            [Quantity, Quantity],
            [[(nx,), (nx + 1,)], [(ny,), (ny + 1,)]],
            ['x_data', 'y_data'])
        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels
        self._set_data(z_data, 'z')
        self.metadata = {} if metadata is None else metadata

    @property
    def z_data(self) -> Quantity:
        return self._z_data*ureg(self._internal_z_data_unit).to(
            self.z_data_unit)

    @z_data.setter
    def z_data(self, value: Quantity) -> None:
        self.z_data_unit = str(value.units)
        self._z_data = value.to(self._internal_z_data_unit).magnitude

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['z_data_unit'])
        super(Spectrum2D, self).__setattr__(name, value)

    def _split_by_indices(self,
                          indices: Union[Sequence[int], np.ndarray]
                          ) -> List[T]:
        """Split data along x-axis at given indices"""
        ranges = self._ranges_from_indices(indices)
        return [type(self)(self.x_data[x0:x1], self.y_data,
                           self.z_data[x0:x1, :],
                           x_tick_labels=self._cut_x_ticks(self.x_tick_labels,
                                                           x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    def broaden(self, x_width: Optional[Quantity] = None,
                y_width: Optional[Quantity] = None, shape: str = 'gauss',
                method: Optional[str] = None) -> T:
        """
        Broaden z_data and return a new broadened Spectrum2D object

        Parameters
        ----------
        x_width
            Scalar float Quantity. The broadening FWHM in x
        y_width
            Scalar float Quantity. The broadening FWHM in y
        shape
            One of {'gauss', 'lorentz'}. The broadening shape
        method
            Can be None or 'convolve'. Currently the only broadening
            method available is convolution with a broadening kernel,
            but this may not produce correct results for unequal bin
            widths. To use convolution anyway, explicitly set
            method='convolve'

        Returns
        -------
        broadened_spectrum
            A new Spectrum2D object with broadened z_data

        Raises
        ------
        ValueError
            If shape is not one of the allowed strings
        ValueError
            If method is None and bins are not of equal size
        """
        bin_centres = [self.get_bin_centres(ax).magnitude for ax in ['x', 'y']]
        bin_widths = [None]*2
        if x_width is not None:
            bin_widths[0] = x_width.to(self.x_data_unit).magnitude
        if y_width is not None:
            bin_widths[1] = y_width.to(self.y_data_unit).magnitude
        z_broadened = self._broaden_data(self.z_data.magnitude, bin_centres,
                                         bin_widths, shape=shape,
                                         method=method)
        return Spectrum2D(
            np.copy(self.x_data.magnitude)*ureg(self.x_data_unit),
            np.copy(self.y_data.magnitude)*ureg(self.y_data_unit),
            z_broadened*ureg(self.z_data_unit), copy.copy(self.x_tick_labels),
            copy.copy(self.metadata))

    def get_bin_edges(self, bin_ax: str = 'x') -> Quantity:
        """
        Get bin edges for the axis specified by bin_ax. If the size of
        bin_ax is one element larger than z_data along that axis, bin_ax
        is assumed to contain bin edges, but if bin_ax is the same size,
        bin_ax is assumed to contain bin centres and a conversion is made.
        In the conversion, the bin edges will not go outside the existing
        data bounds so the first and last bins may be half-size. In addition,
        each bin edge is assumed to be halfway between each bin centre,
        which may not be an accurate assumption in the case of differently
        sized bins.

        Parameters
        ----------
        bin_ax
            The axis to get the bin edges for, 'x' or 'y'
        """
        enum = {'x': 0, 'y': 1}
        bin_data = getattr(self, f'{bin_ax}_data')
        data_ax_len = self.z_data.shape[enum[bin_ax]]
        if self._is_bin_edge(data_ax_len, bin_data.shape[0]):
            return bin_data
        else:
            return self._bin_centres_to_edges(bin_data.magnitude
                                              )*bin_data.units

    def get_bin_centres(self, bin_ax: str = 'x') -> Quantity:
        """
        Get bin centres for the axis specified by bin_ax. If the size of
        bin_ax is the same size as z_data along that axis, bin_ax is
        assumed to contain bin centres, but if bin_ax is one element
        larger, bin_ax is assumed to contain bin centres and a conversion
        is made. In this conversion, the bin centres are assumed to be in
        the middle of each bin, which may not be an accurate assumption in
        the case of differently sized bins.

        Parameters
        ----------
        bin_ax
            The axis to get the bin centres for, 'x' or 'y'
        """
        enum = {'x': 0, 'y': 1}
        bin_data = getattr(self, f'{bin_ax}_data')
        data_ax_len = self.z_data.shape[enum[bin_ax]]
        if self._is_bin_edge(data_ax_len, bin_data.shape[0]):
            return self._bin_edges_to_centres(
                bin_data.magnitude)*bin_data.units
        else:
            return bin_data

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to a dictionary. See Spectrum2D.from_dict for details on
        keys/values

        Returns
        -------
        dict
        """
        return _obj_to_dict(self, ['x_data', 'y_data', 'z_data',
                                   'x_tick_labels', 'metadata'])

    @classmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        """
        Convert a dictionary to a Spectrum2D object

        Parameters
        ----------
        d : dict
            A dictionary with the following keys/values:

            - 'x_data': (n_x_data,) or (n_x_data + 1,) float ndarray
            - 'x_data_unit': str
            - 'y_data': (n_y_data,) or (n_y_data + 1,) float ndarray
            - 'y_data_unit': str
            - 'z_data': (n_x_data, n_y_data) float Quantity
            - 'z_data_unit': str

            There are also the following optional keys:

            - 'x_tick_labels': list of (int, string) tuples
            - 'metadata': dict

        Returns
        -------
        spectrum
        """
        d = _process_dict(d, quantities=['x_data', 'y_data', 'z_data'],
                          optional=['x_tick_labels', 'metadata'])
        return cls(d['x_data'], d['y_data'], d['z_data'],
                   x_tick_labels=d['x_tick_labels'],
                   metadata=d['metadata'])

def apply_kinematic_constraints(spectrum: Spectrum2D,
                                e_i: Quantity = None,
                                e_f: Quantity = None,
                                angle_range: Tuple[float] = (0, 180.)
                                ) -> Spectrum2D:
    """
    Set events to NaN which violate energy/momentum limits:

      - Energy transfer greater than e_i
      - q outside region accessible for given e_i and angle range

    This requires x_data to be in wavevector units and y_data to be energy.

    Either e_i or e_f should be set, according to direct/instrument
    geometry. The other values will be inferred, interpreting y_data as
    energy transfer.

    Parameters
    ----------
    spectrum
        input 2-D spectrum, with |q| x_data and energy y_data
    e_i
        incident energy of direct-geometry spectrometer
    e_f
        final energy of indirect-geometry spectrometer
    angle_range
        min and max scattering angles (2θ) of detector bank in degrees

    Returns
    -------
    Masked spectrum with inaccessible bins set to NaN in z_data.
    """
    try:
        (1 * spectrum.x_data.units).to('1/angstrom')
    except DimensionalityError as error:
        raise ValueError(
            "x_data needs to have wavevector units (i.e. 1/length)"
            ) from error
    try:
        (1 * spectrum.y_data.units).to('eV', 'spectroscopy')
    except DimensionalityError as error:
        raise ValueError(
            "y_data needs to have energy (or wavenumber) units"
            ) from error

    momentum2_to_energy = 0.5 * (ureg('hbar^2 / neutron_mass')
                                 .to('meV angstrom^2'))

    if (e_i is None) == (e_f is None):
        raise ValueError("Exactly one of e_i and e_f should be set. "
                         "(The other value will be derived from energy "
                         "transfer).")

    if e_i is None:   # Indirect geometry: final energy is fixed,
                        # incident energy range is unlimited
        e_f = e_f.to('meV')
        e_i = (spectrum.get_bin_centres(bin_ax='y').to('meV') + e_f)
    elif e_f is None:   # Direct geometry: incident energy is fixed,
                        # max energy transfer = e_i
        e_i = e_i.to('meV')
        e_f = e_i - spectrum.get_bin_centres(bin_ax='y').to('meV')

    k2_i = (e_i / momentum2_to_energy)
    k2_f = (e_f / momentum2_to_energy)

    cos_values = np.asarray(
        _get_cos_range(np.asarray(angle_range) * np.pi / 180.))

    # Momentum goes negative where final energy greater than incident
    # energy; detect this as complex component and set extreme q-bounds to
    # enforce conservation of energy

    # (Complex number sqrt separated from units sqrt for compatibility with
    # older library versions; in newer versions this is not necessary.)
    q_bounds = np.sqrt(k2_i + k2_f
                       - 2 * cos_values[:, np.newaxis]
                           * np.sqrt(k2_i.magnitude * k2_f.magnitude,
                                     dtype=complex)
                           * (k2_i.units * k2_f.units)**0.5
                       )
    q_bounds.magnitude.T[np.any(q_bounds.imag, axis=0)] = [float('Inf'),
                                                           float('-Inf')]
    q_bounds = q_bounds.real

    new_z_data = np.copy(spectrum.z_data.magnitude)
    mask = np.logical_or((spectrum.get_bin_centres(bin_ax='x')[:, np.newaxis]
                          < q_bounds[0][np.newaxis, :]),
                         (spectrum.get_bin_centres(bin_ax='x')[:, np.newaxis]
                          > q_bounds[1][np.newaxis, :]))

    new_z_data[mask] = float('nan')

    return Spectrum2D(
        np.copy(spectrum.x_data.magnitude) * ureg(spectrum.x_data_unit),
        np.copy(spectrum.y_data.magnitude) * ureg(spectrum.y_data_unit),
        new_z_data * ureg(spectrum.z_data_unit),
        copy.copy(spectrum.x_tick_labels),
        copy.deepcopy(spectrum.metadata))


def _get_cos_range(angle_range: Tuple[float]) -> Tuple[float]:
    """
    Get max and min of cosine function over angle range

    These will either be the cosines of the input angles, or, in the case that
    a cosine max/min point lies within the angle range, 1/-1 respectively.

    Method: the angle range is translated such that it starts within 0-2π; then
    we check for the presence of turning points at π and 2π.
    """
    limiting_values = np.cos(angle_range).tolist()

    shift, lower_angle = divmod(min(angle_range), np.pi * 2)
    upper_angle = max(angle_range) - (shift * 2 * np.pi)
    if lower_angle < np.pi < upper_angle:
        limiting_values.append(-1.)
    if lower_angle < 2 * np.pi < upper_angle:
        limiting_values.append(1.)
    return max(limiting_values), min(limiting_values)


def _lorentzian(x: np.ndarray, gamma: float) -> np.ndarray:
    return gamma/(2*math.pi*(np.square(x) + (gamma/2)**2))


def _get_dist_bins(bins: np.ndarray, fwhm: float, extent: float
                   ) -> np.ndarray:
    # Ensure nbins is always odd, and each bin has the same approx width
    # as original x/ybins
    bin_width = np.mean(np.diff(bins))
    nbins = int(np.ceil(2*extent*fwhm/bin_width)/2)*2 + 1
    width = extent*fwhm
    # Prevent xbins from being too large. If user accidentally selects a
    # very large broadening, xwidth and therefore xbins could be
    # extremely large. But for most cases the original nxbins should be
    # smaller
    if nbins > len(bins):
        nbins = int(len(bins)/2)*2 + 1
        width = (bins[-1] - bins[0])/2
    return np.linspace(-width, width, nbins)


def _distribution_1d(xbins: np.ndarray, xwidth: float, shape: str = 'lorentz',
                     extent: float = 3.0) -> np.ndarray:
    x = _get_dist_bins(xbins, xwidth, extent)
    if shape == 'lorentz':
        dist = _lorentzian(x, xwidth)
    dist = dist/np.sum(dist)  # Naively normalise
    return dist
